"""
This example shows how to attach 'done' and 'error' callbacks to a slow
method call.
"""

#### Imports ####

from traits.api import HasTraits
from jigna.api import HTMLWidget, Template
from jigna.qt import QtGui
import time

#### Domain model ####

class Worker(HasTraits):

    def do_work(self):
        """ Simulate doing some work by sleeping a bit :) """
        time.sleep(1)

    def do_illegal_work(self):
        """ Simulate doing an illegal work that will raise an error """
        time.sleep(1)
        raise Exception("Illegal operation")

#### UI layer ####

template = Template(html_file='success_error_callbacks.html')

#### Entry point ####

def main():
    # Start the Qt application
    app = QtGui.QApplication([])

    # Instantiate the domain models
    worker = Worker()

    # Create the jigna based HTML widget which renders the given HTML template
    # with the given context.
    widget = HTMLWidget(template=template, context={'worker': worker})
    widget.show()

    # Start the event loop.
    #
    # The view related code is such that clicking on the buttons in the UI will
    # call methods on the domain model and do something when the method call
    # succeeded or failed.
    app.exec_()

if __name__ == '__main__':
    main()

#### EOF ######################################################################



#
# Enthought product code
#
# (C) Copyright 2013 Enthought, Inc., Austin, TX
# All right reserved.
#

""" Qt implementations of the Jigna Server and Bridge. """


# Standard library.
import json
import os
from os.path import abspath, dirname, join

# Enthought library.
from traits.api import Any, Bool, Str, Instance
from traits.trait_notifiers import set_ui_handler

# Jigna library.
from jigna.core.proxy_qwebview import ProxyQWebView
from jigna.core.wsgi import FileLoader
from jigna.server import Bridge, Server
from jigna.qt import QtWebKit
from jigna.utils.gui import ui_handler

#: Path to jigna.js file
JIGNA_JS_FILE = join(abspath(dirname(__file__)), 'js', 'dist', 'jigna.js')
JIGNA_VUE_JS_FILE = join(
    abspath(dirname(__file__)), 'js', 'dist', 'jigna-vue.js'
)


class QtBridge(Bridge):
    """ Qt (via QWebkit) bridge implementation. """

    #### 'Bridge' protocol ####################################################

    def send_event(self, event):
        """ Send an event. """

        try:
            jsonized_event = json.dumps(event)
        except TypeError:
            return

        if self.webview is None:
            raise RuntimeError("WebView does not exist")

        else:
            # This looks weird but this is how we fake an event being
            # 'received' on the client side when using the Qt bridge!
            self.webview.execute_js(
                'jigna.client.bridge.handle_event(%r);' % jsonized_event
            )

        return

    #### 'QtBridge' protocol ##################################################

    #: The 'WebViewContainer' that contains the QtWebKit malarky.
    webview = Any


class QtServer(Server):
    """ Qt (via QWebkit) server implementation. """

    #### 'Server' protocol ####################################################

    def __init__(self, **traits):
        """ Initialize the Qt server.

        This simply configures the widget to serve the Python model.
        
        """

        # This statement makes sure that when we dispatch traits events on the
        # 'ui' thread, it passes on those events through the Qt layer.
        set_ui_handler(ui_handler)

        super(QtServer, self).__init__(**traits)
        self.webview.setUrl(self.home_url)
        self._enable_qwidget_embedding()

        return

    #: Whether to switch on the debug flag or not.
    #:
    #: In debug mode, we show the QtWebkit inspect tools.
    debug = Bool(False)

    #: The trait change dispatch mechanism to use when traits change.
    trait_change_dispatch = Str('ui')

    def shutdown(self):
        """ Shutdown the server.

        Overridden to make sure we close up the QWebView.

        """

        super(QtServer, self).shutdown()

        self.webview.close()

    ### 'QtServer' protocol ##################################################

    #: The `ProxyQWebView` object which specifies rules about how to handle
    #: different requests etc.
    webview = Instance(ProxyQWebView)
    def _webview_default(self):
        user_root, index_file = self.home_url.split('/')[-2:]

        return ProxyQWebView(
            python_namespace = 'qt_bridge',
            callbacks        = [('handle_request', self.handle_request)],
            debug            = self.debug,
            hosts            = {
                user_root: FileLoader(
                    root      = abspath(self.base_url),
                    overrides = {
                        index_file: self.html,
                        join('jigna', 'jigna.js'): open(JIGNA_JS_FILE).read(),
                        join('jigna', 'jigna-vue.js'):
                        open(JIGNA_VUE_JS_FILE).read()
                    }
                ),
                'root.filesystem': FileLoader(root=abspath(os.sep))
            }
        )

    #### Private protocol #####################################################

    _bridge = Instance(QtBridge)
    def __bridge_default(self):
        return QtBridge(webview=self.webview)

    _plugin_factory = Instance('QtWebPluginFactory')

    def _enable_qwidget_embedding(self):
        """ Allow generic qwidgets to be embedded in the generated QWebView.
        """
        global_settings = QtWebKit.QWebSettings.globalSettings()
        global_settings.setAttribute(QtWebKit.QWebSettings.PluginsEnabled,True)

        self._plugin_factory = QtWebPluginFactory(context=self.context)
        self.webview.page().setPluginFactory(self._plugin_factory)


class QtWebPluginFactory(QtWebKit.QWebPluginFactory):

    MIME_TYPE = 'application/x-qwidget'

    def __init__(self, context):
        self.context = context
        super(self.__class__, self).__init__()

    def plugins(self):
        plugin = QtWebKit.QWebPluginFactory.Plugin()
        plugin.name = 'QWidget'
        mimeType = QtWebKit.QWebPluginFactory.MimeType()
        mimeType.name = self.MIME_TYPE
        plugin.mimeTypes = [mimeType]

        return [plugin]

    def create(self, mimeType, url, argNames, argVals):
        """ Return the QWidget to be embedded.
        """
        if mimeType != self.MIME_TYPE:
            return

        args = dict(zip(argNames, argVals))
        widget_factory = eval(args.get('widget-factory'), self.context)

        return widget_factory()

#### EOF ######################################################################



"""
This example shows how to use Jigna alongwith a custom angularjs application. We
would need a custom angularjs application to add some view related logic.
"""

#### Imports ####
from __future__ import print_function

from jigna.api import HTMLWidget, Template
from jigna.qt import QtGui
import time
from traits.api import HasTraits, Enum, Int

#### Domain model ####

class StopWatch(HasTraits):
    state = Enum('stopped', 'running')
    time = Int

    def start(self):
        self.state = 'running'
        while self.state == 'running':
            time.sleep(1)
            self.time += 1

    def stop(self):
        self.state = 'stopped'

    def reset(self):
        self.time = 0

#### UI layer ####

template = Template(html_file='ex17_custom_angular_application.html')

#### Entry point ####

def main():
    # Start the Qt application
    app = QtGui.QApplication([])

    # Instantiate the domain model
    stop_watch = StopWatch()

    # Create the jigna based HTML widget which renders the given HTML template
    # with the given context.
    #
    # The operations on the stop watch can be controlled from the UI. The view
    # related logic is such that it always displays the integer time of the
    # domain model in a proper hh:mm:ss format.
    widget = HTMLWidget(template=template, context={'stop_watch': stop_watch})
    widget.show()

    # Start the event loop
    app.exec_()

    # Check the values after the UI is closed
    print(stop_watch.time, "seconds")

if __name__ == "__main__":
    main()

#### EOF ######################################################################



#
# Enthought product code
#
# (C) Copyright 2013 Enthought, Inc., Austin, TX
# All right reserved.
#

""" General utility methods for the web version """

import socket

def get_free_port():
    """
    Returns a free socket port. It works by creating an empty socket, binding it
    to port 0 so that the OS automatically assigns a free port to it, obtaining
    the port using `getsockname` and then immediately closing it.

    The application intending to use this port should bind to it immediately so
    that no other application binds to it before us.
    """
    sock = socket.socket()

    # bind to a random port (so that the OS automatically assigns us a free port)
    sock.bind(('', 0))

    # obtain the random port value
    port = sock.getsockname()[1]

    # close the socket so that the port gets free
    sock.close()

    return port


def start_web_app(template, context, port=8000):
    """
    Start a web app at the given port for serving the jigna view for the given
    template and context.
    """
    from tornado.ioloop import IOLoop
    from jigna.web_app import WebApp

    ioloop = IOLoop.instance()

    app = WebApp(template=template, context=context)
    app.listen(port)

    print('Starting the web app on port %s ...' % port)
    ioloop.start()



from __future__ import absolute_import
import sys
import unittest
from unittest import skipIf

from .test_jigna_qt import sleep_while
from .test_jigna_web import TestJignaWebSync, Person

class TestJignaWebAsync(TestJignaWebSync):
    @classmethod
    def setUpClass(cls):
        super(TestJignaWebAsync, cls).setUpClass(async=True)

    def test_callable(self):
        fred = self.fred
        wilma = Person(name='Wilma', age=40)
        self.fred.spouse = wilma
        self.execute_js("var x; jigna.models.model.method('hello').done(function(r){x=r;}); return x;")
        self.wait_and_assert(lambda: fred.called_with != "hello")

        self.execute_js("var x; jigna.models.model.method(1).done(function(r){x=r;}); return x;")
        self.wait_and_assert(lambda: fred.called_with != 1)

        self.execute_js("var x; jigna.models.model.method(10.0).done(function(r){x=r;}); return x;")
        self.wait_and_assert(lambda: fred.called_with != 10.0)

        self.execute_js("var x; jigna.models.model.method([1,2]).done(function(r){x=r;}); return x;")
        self.wait_and_assert(lambda: fred.called_with != [1,2])

        self.execute_js("var x; jigna.models.model.method(jigna.models.model.spouse).done(function(r){x=r;}); return x;")
        self.wait_and_assert(lambda: fred.called_with != wilma)


    @skipIf(sys.platform.startswith('linux'), "Fails on Linux")
    def test_list_sortable(self):
        super(TestJignaWebAsync, self).test_list_sortable()


# Delete this so running just this file does not run all the tests.
del TestJignaWebSync

if __name__ == "__main__":
    unittest.main()



"""
This example demonstrates Jigna's ability to respond to `Event` trait firings.
"""

#### Imports ####

from traits.api import HasTraits, Str, Event, List
from jigna.api import HTMLWidget, Template
from jigna.qt import QtGui
import time

#### Domain model ####

class Downloader(HasTraits):

    file_urls = List(Str)
    file_downloaded = Event

    def download_files(self):
        """ Download the files specified by the file_urls. Fire a
        `file_downloaded` event with each file download. """

        for file_url in self.file_urls:
            # Simulate downloading the file
            time.sleep(1)

            # This will fire the `file_downloaded` event and make `file_url` as
            # the event payload
            self.file_downloaded = file_url

#### UI layer ####

body_html = """
    <div>
         <!-- Start the download of files in a thread -->
         <button ng-click="jigna.threaded(downloader, 'download_files')">
            Download files
         </button>

         <!-- Custom javascript to attach event handlers -->
         <script type='text/javascript'>

            // Show a "Download complete" dialog with every file download.
            var on_file_downloaded = function(event){
                var file_url = event.data.value;
                alert("Download complete: " + file_url);
            };

            // A proxy for the Python model `downloader` is available inside
            // jigna.models
            var downloader = jigna.models['downloader'];

            // Attach the event listener
            jigna.add_listener(downloader, 'file_downloaded', on_file_downloaded)

         </script>
    </div>
"""

template = Template(body_html=body_html)

#### Entry point ####

def main():
    # Start the Qt application
    app = QtGui.QApplication([])

    # Instantiate the domain model
    file_urls = ['images/lena.png', 'videos/big-buck-bunny.mp4']
    downloader = Downloader(file_urls=file_urls)

    # Create the jigna based HTML widget which renders the given HTML template
    # with the given context.
    widget = HTMLWidget(template=template, context={'downloader':downloader})
    widget.show()

    # Start the event loop.
    #
    # Clicking on the button in the UI will call the `download_files` method in
    # a thread. After each file is "downloaded", a Javascript alert should come
    # up. This is in response to a Python event on each file download.
    app.exec_()

if __name__ == "__main__":
    main()

#### EOF ######################################################################



"""
Utility methods for the IPython notebook
"""

from jigna.api import WebApp
from jigna.utils.web import get_free_port
from IPython.display import HTML

def display_jigna(context, template, size=None):
    """
    A `display_html` style method to show rich jigna display for the objects
    within the context.
    """
    PORT = get_free_port()

    app = WebApp(context=context, template=template, port=PORT)
    application = app.create_application()
    application.listen(app.port)

    width, height = size or template.recommended_size
    html = ('<iframe src="http://localhost:%s" width=%s height=%s></iframe>'
            % (PORT, width, height))

    return HTML(html)



"""
This example shows how to embed a jigna view to an existing QWidget framework.
"""

#### Imports ####

from jigna.api import HTMLWidget, Template
from jigna.qt import QtGui
from traits.api import HasTraits, Int, Str

#### Domain model ####

class Employee(HasTraits):
    name = Str
    salary = Int

    def update_salary(self):
        self.salary += int(0.2*self.salary)

#### UI layer ####

body_html = """
    <div>
      Employee name is {{employee.name}}<br/>
      Salary is ${{employee.salary}}<br/>

      <button ng-click='employee.update_salary()'>
        Update salary
      </button>
    </div>
"""

template = Template(body_html=body_html)

def create_jigna_widget():
    """ Create a jigna based HTML widget. This widget is embedded in a QtGui
    application.

    """
    # Instantiate the domain model
    tom = Employee(name='Tom', salary=2000)

    # Create the jigna based HTML widget which renders the given HTML template
    # with the given context.
    return HTMLWidget(template=template, context={'employee': tom})

#### Entry point ####

def main():
    """ Create and show a QMainWindow which embeds the given widget on a button
    click.

    """
    app = QtGui.QApplication([])

    # Define a new QMainWindow
    window = QtGui.QMainWindow()
    window.setMinimumSize(600, 400)

    # Set up the QLayout
    layout = QtGui.QVBoxLayout()
    window.setCentralWidget(QtGui.QWidget())
    window.centralWidget().setLayout(layout)

    # Add a button to the layout which embeds the supplied widget on click.
    button = QtGui.QPushButton("I'm a QPushButton. Press me to embed a widget")
    button.clicked.connect(lambda : layout.addWidget(create_jigna_widget()))
    layout.addWidget(button)

    # Show the window
    window.show()

    app.exec_()

if __name__ == "__main__":
    main()

#### EOF ######################################################################



""" Simple minimal compatibility code for using PyQt4 and PySide
interchangeably only to the extent needed by jigna.

"""


def load_pyside():
    global QtCore, QtGui, QtNetwork, QtWebKit
    from PySide import QtCore, QtGui, QtNetwork, QtWebKit


def load_pyqt():
    global QtCore, QtGui, QtNetwork, QtWebKit

    import sip
    sip.setapi('QDate', 2)
    sip.setapi('QDateTime', 2)
    sip.setapi('QString', 2)
    sip.setapi('QTextStream', 2)
    sip.setapi('QTime', 2)
    sip.setapi('QUrl', 2)
    sip.setapi('QVariant', 2)

    from PyQt4 import QtCore, QtGui, QtNetwork, QtWebKit, Qt

    QtCore.Property = QtCore.pyqtProperty
    QtCore.Signal = QtCore.pyqtSignal
    QtCore.Slot = QtCore.pyqtSlot


def main():

    import os
    import sys

    if os.environ.get('QT_API') == 'pyside' or 'PySide' in sys.modules:
        load_pyside()
    elif os.environ.get('QT_API') == 'pyqt' or 'PyQt4' in sys.modules:
        load_pyqt()
    else:
        try:
            load_pyqt()
        except ImportError:
            load_pyside()

main()




"""
This example shows how to serve a simple jigna view over the web.
"""

#### Imports ####
from __future__ import print_function

from tornado.ioloop import IOLoop
from traits.api import HasTraits, Int, Str, Instance, List
from jigna.api import Template, WebApp

#### Domain model ####

class Person(HasTraits):
    name = Str
    age  = Int

    friends = List(Instance('Person'))

    def add_friend(self):
        self.friends.append(Person(name='Person', age=0))

    #def _anytrait_changed(self, trait, new, old):
    #    print(trait, old, new)

#### UI layer ####

body_html = """
    <div>
      Name: <input ng-model="person.name"><br>
      Age: <input ng-model="person.age" number>
    </div>
    <div style="height:150px; width: 400px; overflow:scroll;" class="resizable">
      <div ng-repeat="friend in person.friends">
         Name: <input ng-model="friend.name"> Age: <input ng-model="friend.age" type="number">
      </div>
    </div>
    <div ng-click="person.add_friend()">+</div>
"""

template = Template(body_html=body_html)

#### Entry point ####

def main():
    # Start the tornado ioloop application
    ioloop = IOLoop.instance()

    # Instantiate the domain model
    fred = Person(name='Fred', age=42)
    N = 1000
    fred.friends = [Person(name=str(i), age=i) for i in range(N)]

    # Create a web app serving the view with the domain model added to its
    # context.
    app = WebApp(template=template, context={'person': fred})
    app.listen(8000)

    # Start serving the web app on port 8000.
    #
    # Point your web browser to http://localhost:8000/ to connect to this jigna
    # web app. Any operation performed on the client directly update the
    # model attributes on the server.
    print('Serving on port 8000...')
    ioloop.start()

if __name__ == "__main__":
    main()

#### EOF ######################################################################



try:
    from __builtin__ import basestring
except ImportError:
    from builtins import str as basestring

from jigna.api import HTMLWidget, Template
from jigna.qt import QtWebKit, QtGui

from traits.api import HasTraits, Str, Int

import unittest

class MyModel(HasTraits):
    attr1 = Str
    attr2 = Int

class TestHTMLWidget(unittest.TestCase):

    def setUp(self):
        self.model = MyModel(attr1="Attr1", attr2=2)
        self.template = Template(body_html="")

        self.app = QtGui.QApplication.instance() or QtGui.QApplication([])

    def test_widget_is_created(self):
        # Create a widget
        widget = HTMLWidget(
            template=self.template, context={'model': self.model}
        )

        # Check if a qwebview widget was created
        self.assertIsNotNone(widget)
        self.assertIsInstance(widget, QtGui.QWidget)
        self.assertIsInstance(widget.webview, QtWebKit.QWebView)

    def test_created_widget_loads_html(self):
        # Create a widget
        widget = HTMLWidget(
            template=self.template, context={'model': self.model}
        )

        # Check if the widget has html loaded in it
        webview = widget.webview
        frame = webview.page().mainFrame()
        self.assertIsInstance(frame.toHtml(), basestring)
        self.assertGreater(len(frame.toHtml()), 0)

    def test_created_widget_can_access_jigna(self):
        # Create a widget
        widget = HTMLWidget(
            template=self.template, context={'model': self.model}
        )

        # Check if `jigna` was loaded on the JS side
        jigna = widget.execute_js("jigna");
        self.assertIsNotNone(jigna)

    def test_jigna_was_initialized(self):
        # Create a widget
        widget = HTMLWidget(
            template=self.template, context={'model': self.model}
        )

        # Check if `jigna` was initialised on the JS side
        client = widget.execute_js("jigna.client");
        self.assertIsNotNone(client)

    def test_jigna_was_initialized_with_python_models(self):
        # Create a widget
        widget = HTMLWidget(
            template=self.template, context={'model': self.model}
        )

        # Check if `jigna` was initialised with the correct python models
        # (check by making sure that primitive variables are the same)
        attr1 = widget.execute_js("jigna.models.model.attr1")
        attr2 = widget.execute_js("jigna.models.model.attr2")
        self.assertEqual(attr1, self.model.attr1)
        self.assertEqual(attr2, self.model.attr2)

    def test_two_widgets_are_created(self):
        widget1 = HTMLWidget(
            template=self.template, context={'model': self.model}
        )

        widget2 = HTMLWidget(
            template=self.template, context={'model': self.model}
        )

        # Check if `jigna` was initialised with the correct python models
        # (check by making sure that primitive variables are the same)
        attr1 = widget1.execute_js("jigna.models.model.attr1")
        attr2 = widget2.execute_js("jigna.models.model.attr2")
        self.assertEqual(attr1, self.model.attr1)
        self.assertEqual(attr2, self.model.attr2)

    def test_events_are_not_sent_after_widget_is_closed(self):
        # Given
        template = Template(body_html="Attr1: {{model.attr1}}")
        widget = HTMLWidget(template=template, context={'model': self.model})

        # When
        self.model.attr1 = "one"
        widget.close()
        self.model.attr1 = "two"

        # Then
        #
        # After the widget is closed, we should not fire any further object
        # changed events.
        self.assertEqual(widget.execute_js("jigna.models.model.attr1"), "one")

if __name__ == "__main__":
    unittest.main()



"""
This example shows the ability to view generic Python objects (not just
HasTraits objects) in HTML using Jigna. The data binding will only be one-way
in this case, i.e. from the UI to the model.
"""

#### Imports ####
from __future__ import print_function

from jigna.api import HTMLWidget, Template
from jigna.qt import QtGui

#### Domain model ####

class Person(object):
    def __init__(self, name, age):
        self.name = name
        self.age = age

#### UI layer ####

body_html = """
  <div>
    Name: <input ng-model="person.name">
    Age: <input ng-model="person.age" type='number'>
  </div>
"""

template = Template(body_html=body_html)

#### Entry point ####

def main():
    # Start the Qt application
    app = QtGui.QApplication([])

    # Instantiate the domain model
    fred = Person(name='Fred', age=42)

    # Create the jigna based HTML widget which renders the given HTML template
    # with the given context.
    #
    # This will behave as a static page since we don't have the traits
    # machinery here to reflect model updates in the view.
    widget = HTMLWidget(template=template, context={'person': fred})
    widget.show()

    # Start the event loop
    app.exec_()

    # Check the values after the UI is closed
    print(fred.name, fred.age)

if __name__ == "__main__":
    main()

#### EOF ######################################################################



#
# Jigna product code
#
# (C) Copyright 2013 Enthought, Inc., Austin, TX
# All right reserved.
#

# Standard library imports.
try:
    from __builtin__ import unicode
except ImportError:
    from builtins import str as unicode

import logging
import sys
import threading
from io import StringIO

# System library imports.
from jigna.qt import QtCore, QtNetwork

# Logger.
logger = logging.getLogger(__name__)


class ProxyAccessManager(QtNetwork.QNetworkAccessManager):
    """ A QNetworkAccessManager subclass which proxies requests for a set of
    hosts and schemes.
    """
    def __init__(self, root_paths={}, hosts={}):
        """ root_paths: Mapping of root paths to WSGI callables.
            hosts: Mapping of hosts to WSGI callables.
        """
        super(ProxyAccessManager, self).__init__()
        self.root_paths = root_paths
        self.hosts = hosts

    def get_url_handler(self, url):
        """ Returns the WSGI callable to be used for specified url.
        """
        handler = self.hosts.get(url.host())

        if handler is None:
            for root, _handler in self.root_paths.items():
                if url.path().split('/')[1] == root:
                    handler = _handler

        return handler

    def inject(self, webview):
        """ Replace the old QNetworkAccessManager instance with this instance.
        """
        old_manager = webview.page().networkAccessManager()

        self.setCache(old_manager.cache())
        self.setCookieJar(old_manager.cookieJar())
        self.setProxy(old_manager.proxy())
        self.setProxyFactory(old_manager.proxyFactory())
        webview.page().setNetworkAccessManager(self)

    ###########################################################################
    # QNetworkAccessManager interface
    ###########################################################################

    def createRequest(self, operation, request, data):
        """ Create a ProxyReply request if url handler is provided by
        `hosts`, else defer to the original QNetworkAccessManager
        `createRequest` method.
        """
        url = request.url()
        handler = self.get_url_handler(url)
        if handler is not None:
            data_str = data and data.readAll() or ''
            return ProxyReply(self, url, operation, request, data_str, handler)

        # Default case, let superclass handle normal web access
        return super(ProxyAccessManager, self).createRequest(
            operation, request, data
        )


class ProxyReply(QtNetwork.QNetworkReply):
    """ QNetworkReply subclass to send a specific request to local wsgi app.
    """
    def __init__(self, parent, url, operation, req, data, handler):
        """ handler is the wsgi app """

        super(ProxyReply, self).__init__(parent)

        self.setRequest(req)
        self.setOperation(operation)
        self.setUrl(url)

        self.req_data = data
        self.handler = handler

        self.buffer = b''
        self._buflock = threading.Lock()
        self.aborted = False

        self.open(self.ReadOnly)

        self._worker = ProxyReplyWorker(self)
        self._worker.start()

        # Handle synchronous requests (webkit sync ajax requests)
        # req.Attribute.QSynchronousHttpNetworkReply may not be defined for
        # pyside compiled with qt 4.7 but still works with qt 4.8
        # QSynchronousHttpNetworkReply = DownloadBufferAttribute + 1 = 16
        if req.attribute(req.Attribute(16)):
            self._worker.wait()

    ###########################################################################
    # QNetworkReply interface
    ###########################################################################

    def abort(self):
        if not self.aborted:
            self.aborted = True
            self.setError(self.OperationCanceledError,
                          'Request Aborted')

    def bytesAvailable(self):
        return super(ProxyReply, self).bytesAvailable() + len(self.buffer)

    def isSequential(self):
        return True

    def readData(self, maxSize):
        with self._buflock:
            data, self.buffer = self.buffer[:maxSize], self.buffer[maxSize:]
        return data


class ProxyReplyWorker(QtCore.QThread):
    """ Worker thread to fetch urls for QNetworkProxy. """

    # Signals to forward to ProxyReply
    metaDataChanged = QtCore.Signal()
    readyRead = QtCore.Signal()
    finished = QtCore.Signal()

    OPERATIONS = {QtNetwork.QNetworkAccessManager.GetOperation: 'GET',
                  QtNetwork.QNetworkAccessManager.PostOperation: 'POST',}

    def __init__(self, reply, parent=None):
        super(ProxyReplyWorker, self).__init__(parent)
        self.reply = reply
        self.metaDataChanged.connect(self.reply.metaDataChanged)
        self.readyRead.connect(self.reply.readyRead)
        self.finished.connect(self.reply.finished)

    ###########################################################################
    # QThread interface.
    ###########################################################################

    def run(self):
        """ handles the request by acting as a WSGI forwarding server. """
        reply = self.reply
        url = reply.url()
        req = reply.request()

        # WSGI environ variables
        env = {
            'REQUEST_METHOD': self.OPERATIONS[reply.operation()],
            'SCRIPT_NAME': '',
            'PATH_INFO': url.path(),
            'SERVER_NAME': url.host(),
            'SERVER_PORT': '80',
            'SERVER_PROTOCOL': 'HTTP/1.1',
            'QUERY_STRING': str(url.encodedQuery()),
            'wsgi.version': (1, 0),
            'wsgi.url_scheme': url.scheme(),
            'wsgi.input': StringIO(unicode(reply.req_data)),
            'wsgi.errors': sys.stderr,
            'wsgi.multithread': False,
            'wsgi.multiprocess': True,
            'wsgi.run_once': False,
        }

        # Set WSGI HTTP request headers
        for head_name in req.rawHeaderList():
            env_name = 'HTTP_' + head_name.data().decode().replace('-','_').upper()
            head_val = req.rawHeader(head_name)
            env[env_name] = head_val.data()

        try:
            local_buf = []
            local_buf_len = 0
            for read in reply.handler(env, self._start_response):
                if reply.aborted:
                    return
                if not isinstance(read, bytes):
                    read = str(read).encode('utf8')
                local_buf.append(read)
                local_buf_len += len(read)
                if local_buf_len >= 8192:
                    # Do not write to buffer on every read, app is slowed down
                    # due to lock contention
                    with reply._buflock:
                        reply.buffer += b''.join(local_buf)
                    local_buf = []
                    local_buf_len = 0
                    self.readyRead.emit()
            with reply._buflock:
                reply.buffer += b''.join(local_buf)

        except Exception as e:
            if reply.aborted:
                return
            reply.setAttribute(
                QtNetwork.QNetworkRequest.HttpStatusCodeAttribute, 500
            )
            reply.setAttribute(
                QtNetwork.QNetworkRequest.HttpReasonPhraseAttribute,
                'Internal Error'
            )
            with reply._buflock:
                reply.buffer += b'WSGI Proxy "Server" Error.\n' + str(e)
        finally:
            self.readyRead.emit()
            self.finished.emit()

    ###########################################################################
    # Private interface.
    ###########################################################################

    def _start_response(self, status, response_headers):
        """ WSGI start_response callable. """
        code, reason = status.split(' ', 1)
        self.reply.setAttribute(
            QtNetwork.QNetworkRequest.HttpStatusCodeAttribute, int(code)
        )
        self.reply.setAttribute(
            QtNetwork.QNetworkRequest.HttpReasonPhraseAttribute, reason
        )
        for name, value in response_headers:
            self.reply.setRawHeader(name, str(value))

        self.metaDataChanged.emit()



#
# Enthought product code
#
# (C) Copyright 2013-2016 Enthought, Inc., Austin, TX
# All right reserved.
#

from textwrap import dedent

# Enthought library.
from traits.api import Bool, HasTraits, Str, Property, Tuple, Int


#### HTML template ############################################################

class Template(HasTraits):
    """ Encapsulation of the HTML/AngularJS template which can be rendered by
    jigna.
    """

    #### 'Template' protocol ##################################################

    #: Should we use the async client or not.
    #:
    #: Async client presents a deferred API and is useful when you want to have
    #: your View served over the web where you don't want to freeze the browser
    #: during synchronous GET calls from the server.
    #:
    #: NOTE: When you're specifying the full HTML for the View, the option to
    #: start an async client is specified at the Javascript level instead of
    #: here, using the Javascript statement: `jigna.initialize({async: true})`.
    #: In that case, the value of this trait becomes moot.
    async = Bool(False)

    #: The base url for all resources (relative urls are resolved corresponding
    #: to the current working directory).
    base_url = Str

    #: The inner HTML for the *body* of the view's document.
    body_html = Str

    #: The inner HTML for the *head* of the view's document.
    head_html = Str

    #: The file which contains the html. `html_file` takes precedence over
    #: `body_html` and `head_html`.
    html_file = Str

    #: The HTML template used for this.
    html_template = Str

    #: The HTML for the entire document.
    #:
    #: The order of precedence in determining its value is:
    #:  1. Directly specified `html` trait
    #:  2. Read the contents of the file specified by the `html_file` trait
    #:  3. Create the jigna template out of specified `body_html` and `head_html`
    #:     traits
    html = Property(Str)
    _html = Str

    #: Recommended size of the HTML (in a (width, height) format)
    recommended_size = Tuple(Int(600), Int(400))

    def _get_html(self):
        """ Get the default HTML document for the given model. """

        # Return the cached html value if the trait is specified directly
        if len(self._html) > 0:
            return self._html

        # Else, read from the html file if it is specified...
        if len(self.html_file) > 0:
            with open(self.html_file) as f:
                html = f.read()

        # ...otherwise, create the template out of body and head htmls
        else:
            async = 'true' if self.async else 'false'
            html = self.html_template.format(
                body_html = self.body_html,
                head_html = self.head_html,
                async     = async,
            )

        return html

    def _set_html(self, html):
        self._html = html

    def _html_template_default(self):
        return dedent("""
        <html ng-app='jigna'>
          <head>
            <script type="text/javascript" src="/jigna/jigna.js"></script>
            <script type="text/javascript">
                jigna.initialize({{async: {async}}});
            </script>

            {head_html}

          </head>

          <body>
            {body_html}
          </body>

        </html>
        """)



#### Imports ####
from __future__ import print_function

from jigna.api import Template, WebApp
from numpy import linspace, sin, pi
from tornado.ioloop import IOLoop
from traits.api import (
    HasTraits, CInt, Str, Property, Array, Instance, on_trait_change
)

#### Domain model ####

class DomainModel(HasTraits):
    """
    The algorithmic domain model which specifies the mathematical relationship
    between x and y.
    """

    #: Independent variable of the domain equation
    x = Array
    def _x_default(self):
        return linspace(-2*pi, 2*pi, 200)

    #: Dependent variable of the domain equation
    y = Property(Array, depends_on=['x', 'scaling_factor'])
    def _get_y(self):
        return sin(self.scaling_factor * self.x) / self.x

    #: A scaling factor to tune the output
    scaling_factor = CInt

#### Controller layer ####

class PlotController(HasTraits):
    """
    A Controller class which creates a matplotlib plot object (in the form of a
    png image bytestream) for the given domain model.
    """

    #: Instance of the domain model which is being displayed by this controller
    domain_model = Instance(DomainModel)

    #: The 'png' bytestream representation of the matplotlib plot object
    plot = Str

    @on_trait_change('domain_model.scaling_factor')
    def update_plot(self):
        # Use the Agg backend to generate images without making the window appear
        import matplotlib
        matplotlib.use('Agg')

        # Generate the plot
        from matplotlib import pyplot
        pyplot.clf()
        pyplot.plot(self.domain_model.x, self.domain_model.y)

        # Generate image data in png format
        from io import StringIO
        stream = StringIO()
        pyplot.savefig(stream, format='png')
        stream.seek(0)
        self.plot = stream.buf.encode('base64')

#### UI layer ####

body_html = """
   <div>
        Scaling factor: <input type="range" ng-model="domain_model.scaling_factor"
                        min=0 max=30><br>
        Plot:<br>
        <img ng-src="data:image/png;base64,{{plot_controller.plot}}">
    </div>
"""

template = Template(body_html=body_html)

#### Entry point ####

def main():
    # Start the tornado ioloop application
    ioloop = IOLoop.instance()

    # Instantiate the domain model and the plot controller
    domain_model = DomainModel(scaling_factor=15)
    plot_controller = PlotController(domain_model=domain_model)

    # Create a WebApp to render the HTML template with the given context.
    #
    # The web app shows a matplotlib plot in a browser by displaying the png
    # image of the plot. Moving the HTML slider on the UI changes the domain
    # model and hence the plot image.
    app = WebApp(
        template=template,
        context={
            'domain_model': domain_model,
            'plot_controller': plot_controller
        }
    )
    app.listen(8000)

    # Start serving the web app on port 8000.
    print('Serving on port 8000...')
    ioloop.start()

if __name__ == "__main__":
    main()

#### EOF ######################################################################



from .template import Template
from .vue_template import VueTemplate
from .core.concurrent import Future
from .html_widget import HTMLWidget

# Wrapping the WebApp import so that you can use jigna even if you don't have
# tornado install
try:
    from .web_app import WebApp
except ImportError:
    pass



"""
This example demonstrates Jigna's ability to call public methods of the
traits model from the HTML interface. You can supply primitive arguments and
also pass instances.
"""

#### Imports ####
from __future__ import print_function

from traits.api import HasTraits, Int, Str, Instance
from jigna.api import HTMLWidget, Template
from jigna.qt import QtGui

#### Domain model ####

class Person(HasTraits):
    name = Str
    age = Int
    spouse = Instance('Person')

    def greet(self):
        """ Simple method without any arguments. """
        print('Greetings %s, from the Javascript world!' % self.name)

    def update_age(self, age):
        """ Method which takes a primitive argument. """
        self.age = age

    def marry(self, person):
        """ Method which takes an instance argument. """
        self.spouse = person
        person.spouse = self

#### UI layer ####

body_html = """
    <div>
      Name: <input ng-model="person.name"> <br/>
      Age: {{person.age}} <br/>
      Spouse: {{person.spouse.name}} <br/>

      <button ng-click="person.greet()">Greet!</button>
      <button ng-click="person.update_age(25)">Grow old</button>
      <button ng-click="person.marry(spouse)">Marry</button>
    </div>
"""

template = Template(body_html=body_html)

#### Entry point ####

def main():
    # Start the Qt application
    app = QtGui.QApplication([])

    # Instantiate the domain models
    fred = Person(name='Fred', age=14)
    wilma = Person(name='Wilma', age=25)

    # Create the jigna based HTML widget which renders the given HTML template
    # with the given context.
    widget = HTMLWidget(
        template=template, context={'person': fred, 'spouse': wilma}
    )
    widget.show()

    # Start the event loop.
    #
    # Clicking on the buttons in the UI will make blocking calls to the
    # corresponding methods on the domain model. You can supply primitive as
    # well as instance objects as arguments of the method.
    app.exec_()

    # Check the final values after the UI is closed
    print(fred.name, fred.age, fred.spouse.name)

if __name__ == '__main__':
    main()

#### EOF ######################################################################



"""
This example shows how to call slow methods of the application model such
that the UI doesn't block while the method is executing. We present an API to
call such methods in a thread.

Note: It's the user's responsibility to make sure that there aren't thread
safety related issues with that method.
"""

#### Imports ####
from __future__ import print_function

from traits.api import HasTraits, Int, Str, Instance
from jigna.api import HTMLWidget, Template
from jigna.qt import QtGui
import time

#### Domain model ####

class Package(HasTraits):
    name = Str
    version = Str

class Installer(HasTraits):
    current = Instance('Package')
    progress = Int

    def install(self, package):
        self.current = package
        while self.progress < 100:
            time.sleep(0.5)
            self.progress += 10

#### UI layer ####

body_html = """
    <div>
        <!-- Show a button initially to install in a thread -->
        <button ng-show="installer.progress==0"
                ng-click="jigna.threaded(installer, 'install', new_package)">
            Install {{new_package.name}}-{{new_package.version}}
        </button>

        <!-- Show an "Installing..." text while progress is between 0 and 100 -->
        <div ng-show="installer.progress > 0 && installer.progress < 100">
            Installing {{installer.current.name}}-{{installer.current.version}}...
        </div>

        <!-- Show "Done" when the progress is 100% -->
        <div ng-show="installer.progress >= 100">
            Done!
        </div>

        <!-- A graphical progress bar -->
        <div class='progress-bar-container'>
            <div class='completed-progress'
                 ng-style="{ width: installer.progress + '%' }">
            </div>
        </div>
    </div>

    <style type="text/css">
        .progress-bar-container {
            height: 10px;
            border: solid 1px gray;
            margin-top: 10px;
        }
        .completed-progress {
            background-color: blue;
            height: 100%;
        }
    </style>
"""

template = Template(body_html=body_html)

#### Entry point ####

def main():
    # Start the Qt application
    app = QtGui.QApplication([])

    # Instantiate the domain models
    installer = Installer()
    pandas = Package(name='Pandas', version='1.0')

    # Create the jigna based HTML widget which renders the given HTML template
    # with the given context.
    widget = HTMLWidget(
        template=template,
        context={'installer': installer, 'new_package': pandas}
    )
    widget.show()

    # Start the event loop.
    #
    # Clicking on the button in the UI will call the `install` method in a
    # thread so that the UI is still responsive while the method is executing.
    # The progress bar is also updated as the method progresses.
    app.exec_()

    # Check the final values
    print(installer.current.name, installer.current.version)

if __name__ == '__main__':
    main()

#### EOF ######################################################################



from __future__ import absolute_import

import unittest

from .test_jigna_qt import Person, TestJignaQt, AddressBook

body_vue_html = """
    <div>
      Name: <input v-model="model.name">
      Age: <input v-model="model.age" number>
      <br/>
      Fruits:
      <ul>
        <li v-for="fruit in model.fruits" track-by="$index">
          <input v-model="fruit">
        </li>
      </ul>

      <br/>

      Friends:
      <ul>
        <li v-for="friend in model.friends">
          Name: <input v-model="friend.name">
          Age: <input v-model="friend.age" number>
          Fruits:
          <ul>
            <li v-for="fruit in friend.fruits">
              <input v-model="fruit">
            </li>
          </ul>
        </li>
      </ul>
    </div>

    Spouse: <br/>
    Name: {{model.spouse.name}} Age: {{model.spouse.age}}

   <h3>Addressbook</h3>
   <button v-on:click="threaded(addressbook, 'create')" id="create">
    Create
   </button>
   <br/>
   <ul>
    <li v-for="contact in addressbook.contacts">
     <label>Name:</label> <input v-model="contact.name">
     <label>Number:</label> <input v-model="contact.number">
    </li>
   </ul>
"""

class TestJignaVueQt(TestJignaQt):

    @classmethod
    def setUpClass(cls):
        from jigna.api import HTMLWidget, VueTemplate
        from jigna.utils import gui
        from jigna.qt import QtGui
        qapp = QtGui.QApplication.instance() or QtGui.QApplication([])
        template = VueTemplate(body_html=body_vue_html)
        fred = Person(name='Fred', age=42)
        addressbook = AddressBook()
        widget = HTMLWidget(
            template=template,
            context={'model':fred, 'addressbook': addressbook})
        widget.show()
        gui.process_events()
        cls.widget = widget
        cls.fred = fred
        cls.addressbook = addressbook

del TestJignaQt

if __name__ == "__main__":
    unittest.main()



"""
This example shows how to use Jigna using the Vue.js Javascript framework,
instead of AngularJS.
"""

#### Imports ####
from __future__ import print_function

from traits.api import HasTraits, Int, Str, List, Instance
from jigna.vue_template import  VueTemplate
from jigna.web_app import WebApp
from tornado.ioloop import IOLoop


#### Domain model ####

class Person(HasTraits):
    name = Str
    age  = Int
    fruits = List(Str)
    friends = List(Instance('Person'))

    def add_fruit(self, name='fruit'):
        self.fruits.append(name)

    def remove_fruit(self, index):
        del self.fruits[index]

    def update_name(self, name):
        print("Name updated to", name)
        self.name = name

    def add_friend(self):
        self.friends.append(Person(name='Person', age=0))

    def remove_friend(self, index):
        del self.friends[index]

    def _fruits_items_changed(self, l_event):
        print(l_event.added, l_event.removed, l_event.index)

    def _name_changed(self, old, new):
        print("name", old, new)

    def _age_changed(self, old, new):
        print("age", old, new)


#### Entry point ####

def main():
    # Start the tornado ioloop application
    ioloop = IOLoop.instance()

    # Instantiate the domain model

    fred = Person(name='Fred', age=42, fruits=['pear', 'apple'])
    template = VueTemplate(html_file='ex22_vuejs_demo.html')

    # Create the jigna based HTML widget which renders the given HTML template
    # with the given context.
    app = WebApp(template=template, context={'person': fred})
    app.listen(8000)

    # Schedule an update to a model variable after 5 seconds. This update
    # will be reflected in the UI immediately.
    ioloop.call_later(5, setattr, fred, 'update_name', 'Guido')
    ioloop.call_later(5, fred.add_fruit)
    ioloop.call_later(5, fred.add_friend)

    # Start serving the web app on port 8000.
    #
    # Point your web browser to http://localhost:8000/ to connect to this jigna
    # web app. Any operation performed on the client directly update the
    # model attributes on the server.
    print('Serving on port 8000...')
    ioloop.start()


if __name__ == "__main__":
    main()

#### EOF ######################################################################



""" This example shows how to show mayavi in the web version using the webGL
backend for VTK.
"""

#### Imports ####
from __future__ import print_function

import json
import numpy as np
from mayavi import mlab
from mayavi.core.api import PipelineBase
from traits.api import HasTraits, Instance, Int, Str
from tvtk.api import tvtk
from jigna.web_app import WebApp
from jigna.template import Template
from tornado.ioloop import IOLoop

mlab.options.offscreen = True
mlab.options.backend = 'test'

#### Domain Model ####

class MeshData(HasTraits):

    # Data in the format of an old style VTK file.
    filedata = Str

    # The colors to use for the points, in json format.
    colors = Str

    # The type of the mesh, for now only polygons are supported.
    type = Str("POLYGONS")

    @classmethod
    def from_dataset(cls, dataset, module_manager):
        filedata = cls._dataset_to_string(dataset)
        colors = cls._dataset_to_colors(dataset, module_manager)

        return MeshData(filedata=filedata, colors=colors)

    #### Private protocol #####################################################

    @classmethod
    def _dataset_to_string(cls, dataset):
        """
        Convert the dataset to a vtk filedata string.
        """
        writer = tvtk.PolyDataWriter(
            input=dataset,
            write_to_output_string=True
        )
        writer.write()
        return writer.output_string

    @classmethod
    def _dataset_to_colors(cls, dataset, module_manager):
        """
        Given the dataset, extract the colors array in a jsonized format.
        """
        scm = module_manager.scalar_lut_manager
        scalars = dataset.point_data.scalars

        colors = scm.lut.map_scalars(scalars, 0, -1).to_array()/255.0
        colors_on_points = colors[cls._point_indices(dataset)]

        return json.dumps(colors_on_points.tolist())

    @classmethod
    def _point_indices(cls, dataset):
        """
        Given the dataset, obtain the polygon connectivity array and generate
        the indices of the points in the polygon.
        """
        conn = dataset.polys.to_array()
        npoly = conn.size / 4
        choice = np.zeros(npoly*3, dtype=int)
        for start in (1, 2, 3):
            choice[start-1::3] = np.arange(start, npoly*4, step=4)
        return conn[choice]

class Plotter3D(HasTraits):

    #: expression to be visualized
    expression = Str
    def _expression_changed(self, expr):
        if self.plot is None:
            x, y, z = np.mgrid[-5:5:32j, -5:5:32j, -5:5:32j]
        else:
            x, y, z = self.x, self.y, self.z
        g = np.__dict__
        try:
            s = eval(expr, g, dict(x=x, y=y, z=z))
        except:
            pass
        else:
            if self.plot is None:
                self.x, self.y, self.z = x, y, z
                self.plot = mlab.contour3d(x, y, z, s, contours=self.n_contour)
            else:
                self.plot.mlab_source.set(scalars=s)
            self._update_mesh_data()

    #: number of contours for the visualization
    n_contour = Int(2)
    def _n_contour_changed(self, value):
        if 0 < value < 20:
            if self.plot:
                self.plot.contour.number_of_contours = value
                self._update_mesh_data()

    plot = Instance(PipelineBase)

    mesh_data = Instance(MeshData, ())

    def _update_mesh_data(self):
        self.mesh_data.copy_traits(
            MeshData.from_dataset(
                dataset=self.plot.contour.outputs[0],
                module_manager=self.plot.module_manager
            )
        )

#### UI layer ####

template = Template(html_file='ex20_mayavi_webgl_demo.html')

#### Entry point ####

def main():
    # Start the tornado ioloop application
    ioloop = IOLoop.instance()

    # Instantiate the domain model
    plotter = Plotter3D(expression="x*x*0.5 + y*y + z*z*2.0", n_contour=4)

    # Create a web app serving the view with the domain model added to its
    # context.
    app = WebApp(template=template, context={'plotter': plotter})
    app.listen(8000)

    # Start serving the web app on port 8000.
    print('Serving on port 8000...')
    ioloop.start()

if __name__ == '__main__':
    main()



"""
This example shows how to embed a Mayavi QWidget inside the jigna view using
an <object> tag.

Mayavi widget is different than a Chaco widget since you can interact with the
widget so it demonstrates that mouse events are properly forwarded to the
underlying QWidget when you embed it in jigna view.
"""

#### Imports ####

from jigna.api import HTMLWidget, Template
from jigna.qt import QtGui
from mayavi.core.api import PipelineBase
from mayavi.core.ui.api import MlabSceneModel
from traits.api import HasTraits, Instance, CInt, on_trait_change


#### Domain model ####

class SceneController(HasTraits):
    n_meridional    = CInt(10)
    n_longitudinal  = CInt(10)

    scene = Instance(MlabSceneModel, ())

    plot = Instance(PipelineBase)

    # When the scene is activated, or when the parameters are changed, we
    # update the plot.
    @on_trait_change('n_meridional,n_longitudinal,scene.activated')
    def update_plot(self):
        x, y, z, t = self.get_data()
        if self.plot is None:
            self.plot = self.scene.mlab.plot3d(x, y, z, t,
                                tube_radius=0.025, colormap='Spectral')
        else:
            self.plot.mlab_source.set(x=x, y=y, z=z, scalars=t)

    def get_data(self):
        """ Obtain the x,y,z,t data for the domain equation and given values of
        n_meridional and n_longitudinal
        """
        from numpy import arange, pi, cos, sin
        phi = arange(0.0, 2*pi + 0.5*pi/1000, pi/1000, 'd')

        mu = phi*self.n_meridional
        x = cos(mu) * (1 + cos(self.n_longitudinal * mu/self.n_meridional)*0.5)
        y = sin(mu) * (1 + cos(self.n_longitudinal * mu/self.n_meridional)*0.5)
        z = 0.5 * sin(self.n_longitudinal*mu/self.n_meridional)
        t = sin(mu)

        return x, y, z, t

    def create_scene_widget(self):
        """ Factory method to create the QWidget for the Mayavi scene.

        This follows the same approach as the Chaco example i.e. create a hidden
        traitsui view based on the mayavi SceneEditor and return it's 'control'
        to obtain the required QWidget.
        """
        from traitsui.api import View, Item
        from mayavi.core.ui.api import MayaviScene, SceneEditor
        view = View(Item('scene', show_label=False,
                         editor=SceneEditor(scene_class=MayaviScene)),
                         resizable=True)
        ui = self.edit_traits(view=view, parent=None, kind='subpanel')

        return ui.control


#### UI layer ####

body_html = """
    <div>
      N meridonial: <input type="range" ng-model="scene_controller.n_meridional"
                       min=0 max=30><br>
      N longitudinal: <input type="range" ng-model="scene_controller.n_longitudinal"
                       min=0 max=30><br>
      Plot:<br>

      <object type="application/x-qwidget" width="400" height="400"
              widget-factory="scene_controller.create_scene_widget">
      </object>
    </div>
"""

template = Template(body_html=body_html)

#### Entry point ####

def main():
    # Start the Qt application
    app = QtGui.QApplication.instance() or QtGui.QApplication([])

    # Instantiate the domain model
    scene_controller = SceneController()

    # Create the jigna based HTML widget which renders the given HTML template
    # with the given context.
    #
    # The view contains an embedded Mayavi QWidget showing a visualization of
    # the domain model. Moving the sliders on the UI changes the domain model and
    # hence the Mayavi visualization.
    widget = HTMLWidget(
        template=template, context={'scene_controller': scene_controller}
    )
    widget.show()

    # Start the event loop
    app.exec_()

if __name__ == "__main__":
    main()

#### EOF ######################################################################



"""
This example shows how to serve a simple jigna view over the web.
"""

#### Imports ####
from __future__ import print_function

import argparse

from tornado.ioloop import IOLoop
from traits.api import HasTraits, Int, Str, Instance, List, Dict
from jigna.vue_template import VueTemplate
from jigna.template import Template
from jigna.web_app import WebApp

#### Domain model ####


class Person(HasTraits):
    name = Str
    age  = Int
    start = Int
    friends = List(Instance('Person'))
    spouse = Instance('Person')
    fruits = List(Str)
    tags = Dict

    def add_friend(self):
        self.friends.append(Person(name='Person', age=0))

    def remove_friend(self, index):
        del self.friends[index]

    def _anytrait_changed(self, trait, new, old):
        if trait != 'friends':
            print(trait, old, new)


#### UI layer ####

body_html_vue = """
<html>
    <head>
        <!-- Load the jigna-vue script. The location of jigna-vue script file
        will always be '/jigna/jigna-vue.js' -->
        <script type='text/javascript' src='/jigna/jigna-vue.js'></script>

        <!-- Once jigna is loaded, initialize it. -->
        <script type='text/javascript'>
            jigna.initialize({async:%s, debug:true});
        </script>
    </head>

  <body>
    <h2> Using Vue.js with async: %s, N={{person.friends.length}}</h2>
    <div>
      Name: <input v-model="person.name"><br/>
      Age: <input v-model="person.age" number><br/>
      Spouse: {{person.spouse.name}}<br/>
      Start: <input v-model="person.start" number>
      <div v-for="(tag, value) in person.tags">
       <label>{{tag}}</label><input v-model="value">
      </div>
    </div>
    <div style="height:150px; width: 1000px; overflow:scroll;" class="resizable">
      <div v-for="friend in person.friends | limitBy 20 person.start">
         Name: <input v-model="friend.name"> Age: <input v-model="friend.age" number>
         <div v-if="friend.spouse">Spouse: {{friend.spouse.name}}</div>
        <label v-on:click="person.remove_friend(person.start + $index)">-</label>
      </div>
    </div>
    <div v-on:click="person.start = Math.min(person.start + 10, person.friends.length)"> &gt;</div>
    <div v-on:click="person.add_friend()">+</div>


  <script>
      var vm = undefined;

      jigna.ready.done(function() {
          var vm = new Vue({
              el: "body", // Attach to the body tag.
              data: jigna.models,
          });
      });

    </script>

  </body>

"""

body_html_ng = """
<html ng-app='jigna'>
    <head>
        <script type='text/javascript' src='/jigna/jigna.js'></script>

        <!-- Once jigna is loaded, initialize it. -->
        <script type='text/javascript'>
            jigna.initialize({async:%s, debug:true});
        </script>
    </head>

  <body>
    <h2> Using angular.js with async: %s, N={{person.friends.length}}</h2>
    <div>
      Name: <input ng-model="person.name"><br/>
      Age: <input ng-model="person.age" type="number"><br/>
      Spouse: {{person.spouse.name}}<br/>
      Start: <input ng-model="person.start" type="number">
      <div ng-repeat="(tag, value) in person.tags">
       <label>{{tag}}</label><input ng-model="value">
      </div>
    </div>
    <div style="height:150px; width: 1000px; overflow:scroll;" class="resizable">
      <div ng-repeat="friend in person.friends.slice(person.start, person.start+20)">

         Name: <input ng-model="friend.name"> Age: <input ng-model="friend.age" number>
         <div ng-if="friend.spouse">Spouse: {{friend.spouse.name}}</div>
        <label ng-click="person.remove_friend(person.start + $index)">-</label>
      </div>
    </div>
    <div ng-click="person.start = Math.min(person.start + 10, person.friends.length)"> &gt;</div>
    <div ng-click="person.add_friend()">+</div>

  </body>

"""


#### Entry point ####

def main(async=True, N=10000, engine='vue'):
    s_async = 'true' if async else 'false'
    if engine == 'vue':
        template = VueTemplate(
            html=body_html_vue%(s_async, s_async)
        )
    elif engine == 'ng':
        template = Template(
            html=body_html_ng%(s_async, s_async)
        )

    # Start the tornado ioloop application
    ioloop = IOLoop.instance()

    # Instantiate the domain model
    fred = Person(name='Fred', age=42)
    fred.tags = dict(x=1, y=2, z=3)
    fred.spouse = Person(name='Wilma', age=40)
    fred.friends = [Person(name=str(i), age=i) for i in range(N)]

    # Create a web app serving the view with the domain model added to its
    # context.
    import socket
    from jigna.utils.web import get_free_port
    #port = get_free_port()
    port = 8000
    app = WebApp(template=template, context={'person': fred}, async=async)
    #addr = socket.gethostbyname(socket.gethostname())
    #print(addr)
    app.listen(port)

    # Start serving the web app on port 8000.
    #
    # Point your web browser to http://localhost:8000/ to connect to this jigna
    # web app. Any operation performed on the client directly update the
    # model attributes on the server.
    print('Serving on port %s...'%port)
    from threading import Thread
    def set_attrs():
        import time
        time.sleep(5)
        fred.name = 'Guido'
        fred.age = 25
        time.sleep(0.1)
        del fred.tags['z']
        time.sleep(0.1)
        fred.tags['a'] = 1
        time.sleep(0.1)
        fred.tags['a'] = 100

    t = Thread(target=set_attrs)
    t.setDaemon(True)
    t.start()
    ioloop.start()


if __name__ == "__main__":
    parser = argparse.ArgumentParser(
        description='Jigna test code',
        formatter_class=argparse.ArgumentDefaultsHelpFormatter
    )
    parser.add_argument(
        '--sync', help='Use sync mode', action='store_true', default=False
    )
    parser.add_argument(
        '-n', help='Use N instances.', default=50000, type=int
    )
    parser.add_argument(
        '--engine', help='Use specific engine', default='vue',
        choices=['ng', 'vue']
    )
    args = parser.parse_args()
    main(not args.sync, args.n, args.engine)

#### EOF ######################################################################



"""
This example demonstrates how updates to the application model's attributes are
reflected in the HTML view.
"""

#### Imports ####

from traits.api import HasTraits, Str
from jigna.api import HTMLWidget, Template
from jigna.qt import QtGui
from jigna.utils.gui import do_after

#### Domain model ####

class MOTD(HasTraits):
    message = Str

    def update_message(self, message):
        self.message = message

#### UI layer ####

body_html = """
    <div>
      Message of the day: {{motd.message}}
    </div>
"""

template = Template(body_html=body_html)

#### Entry point ####

def main():
    # Start the Qt application
    app = QtGui.QApplication([])

    # Instantiate the domain model
    motd = MOTD(message="Explicit is better than implicit.")

    # Create the jigna based HTML widget which renders the given HTML template
    # with the given context.
    widget = HTMLWidget(template=template, context={'motd':motd})
    widget.show()

    # Schedule an update to a model variable after 2.5 seconds. This update
    # will be reflected in the UI immediately.
    do_after(2500, motd.update_message, "Flat is better than nested.")

    # Start the event loop
    app.exec_()

if __name__ == "__main__":
    main()

#### EOF ######################################################################



#
# Enthought product code
#
# (C) Copyright 2013 Enthought, Inc., Austin, TX
# All right reserved.
#

# Standard Library
import os
from os.path import join

# Tornado Library
from tornado import web

# Local Library
from .web_server import WebServer, AsyncWebServer


class WebApp(web.Application):
    """ A web based App to serve the jigna template with a given context over
    the web where it can be viewed using a regular web browser. """

    def __init__(self, handlers=None, default_host="", transforms=None,
                 context=None, template=None, trait_change_dispatch="same",
                 async=False, **kw):

        if template is not None:
            template.async = async
        self.context = context
        self.template = template
        self.trait_change_dispatch = trait_change_dispatch
        self.async = async

        if handlers is None:
            handlers = []
        handlers += self._create_handlers()

        super(WebApp, self).__init__(handlers, default_host, transforms, **kw)

    #### Private protocol #####################################################

    def _create_handlers(self):
        """
        Create the web application serving the given context. Returns the
        tornado application created.
        """

        # Set up the WebServer to serve the domain models in context
        klass = AsyncWebServer if self.async else WebServer
        server = klass(
            base_url              = join(os.getcwd(), self.template.base_url),
            html                  = self.template.html,
            context               = self.context,
            trait_change_dispatch = self.trait_change_dispatch
        )

        return server.handlers



"""
This example shows two-way data binding for the `List` traits whose items are of
an instance (non-primitive) type.
"""

#### Imports ####
from __future__ import print_function

from traits.api import HasTraits, Instance, Str, List
from jigna.api import HTMLWidget, Template
from jigna.qt import QtGui
from jigna.utils.gui import do_after

#### Domain model ####

class Person(HasTraits):
    name = Str

    friends = List(Instance('Person'))

#### UI layer ####

# Looping over the list of instances. No surprises here.
body_html = """
    <div>
      Name: {{person.name}} <br/>

      Friends:
      <div ng-repeat="friend in person.friends">
        <input ng-model="friend.name">
      </div>
    </div>
"""

template = Template(body_html=body_html)

#### Entry point ####

def main():
    # Start the Qt application
    app = QtGui.QApplication([])

    # Instantiate the domain model
    fred = Person(name='Fred', friends=[Person(name='Dino')])

    # Create the jigna based HTML widget which renders the given HTML template
    # with the given context.
    widget = HTMLWidget(template=template, context={'person':fred})
    widget.show()

    # Schedule some operations on the list.
    #
    # We're trying to append and insert instances in the list in the future.
    # This should be reflected in the UI.
    do_after(2500, fred.friends.append, Person(name='Wilma'))
    do_after(5000, fred.friends.insert, 0, Person(name='Barney'))

    # Start the event loop
    app.exec_()

    # Check the final values of the list attribute
    print([friend.name for friend in fred.friends])

if __name__ == "__main__":
    main()

#### EOF ######################################################################



#
# Enthought product code
#
# (C) Copyright 2013 Enthought, Inc., Austin, TX
# All right reserved.
#
""" A Qt based HTML widget to render a jigna UI. """


from __future__ import absolute_import

# Standard library.
import os
from os.path import join

# Local library.
from .qt import QtCore, QtGui
from .qt_server import QtServer


class HTMLWidget(QtGui.QWidget):
    """ A Qt based HTML widget to render a jigna UI. """

    #### 'object' protocol ####################################################
    
    def __init__(
        self, parent=None, window_flags=QtCore.Qt.Widget, context=None,
        template=None, debug=False
    ):
        """ Constructor. """
        
        super(HTMLWidget, self).__init__(parent, window_flags)

        # Private protocol.
        self._context  = context
        self._template = template
        self._debug    = debug
        self._server   = self._create_server()

        # fixme: This has to be a public attribute for testing *only*.
        self.webview   = self._server.webview
        
        # Set the widget layout.
        self.setLayout(QtGui.QVBoxLayout())
        self.layout().setContentsMargins(0, 0, 0, 0)
        self.layout().addWidget(self._server.webview)
        self.resize(*template.recommended_size)

    def execute_js(self, js):
        """ Execute the given JS string in the HTML widget. """

        return self._server.webview.execute_js(js)

    #### 'QWidget' protocol ###################################################

    def closeEvent(self, event):
        """ Called when there is a request to close the widget. """

        if event.isAccepted():
            self._server.shutdown()

    #### Private protocol #####################################################

    def _create_server(self):
        """ Create a jigna Qt server to serve the domain models. """

        server = QtServer(
            base_url = join(os.getcwd(), self._template.base_url),
            html     = self._template.html,
            context  = self._context,
            debug    = self._debug
        )

        return server



"""
This example shows how to add additional resources like CSS, javascript
and image files in your html by specifying a base url.
"""

#### Imports ####

from traits.api import HasTraits, Int, Str
from jigna.api import HTMLWidget, Template
from jigna.qt import QtGui

#### Domain model ####

class Person(HasTraits):
    name = Str
    age  = Int

#### UI layer ####

body_html = """
    <div>
        <img ng-src='{{person.name}}.png' /><br/>
        <div id="caption" class='red'>
            {{person.name}} - {{person.age}} years old
        </div>
    </div>

    <script type='text/javascript' src='color_change.js'></script>
    <link rel='stylesheet' href='colors.css' />
"""

# The base_url field specifies where to look when trying to get external
# resources(defaults to an empty string, i.e. the current directory)
template = Template(body_html=body_html, base_url='ex5_data/',
    recommended_size=(600, 600)
)

#### Entry point ####

def main():
    # Start the Qt application
    app = QtGui.QApplication([])

    # Instantiate the domain model
    lena = Person(name='Lena', age=28)

    # Create the jigna based HTML widget which renders the given HTML template
    # with the given context.
    widget = HTMLWidget(template=template, context={'person':lena})
    widget.show()

    # Start the event loop.
    #
    # You should see that user resources like CSS, images and custom JS are
    # pulled in properly from the `user_resources_data` directory and displayed
    # in the view.
    app.exec_()

if __name__ == "__main__":
    main()

#### EOF ######################################################################



#
#
# (C) Copyright 2013-2016 Enthought, Inc., Austin, TX
# All right reserved.
#

"""Module to support asynchronous execution of code."""

# System library imports.
try:
    from __builtin__ import unicode as utext
except ImportError:
    from builtins import str as utext

import sys
from threading import Thread, RLock
from functools import wraps

# Enthought library imports.
from traits.api import (HasTraits, Any, Range, Undefined, Instance, Str,
    Property, Enum, ReadOnly, DelegatesTo, Event)


def set_trait_later(obj, trait, value):
    from ..utils import gui
    gui.set_trait_later(obj, trait, value)


################################################################################
# `Signal` class.
################################################################################

class Signal(HasTraits):
    """ A very simple object signal emitted with only one argument.

    Very similar (and inspired by) Qt's Signals, except that this is object
    based signal instead of class based descriptor signals, i.e., you don't
    need to define signals in the class body, you can directly use Signal
    by instantiating a Signal object.

    Usage:
    ------

        >>> s = Signal()
        >>> def func(arg):
        ...     print 'emitted:', arg
        >>> s.connect(func)
        >>> s.emit('value')
        emitted: value
        >>> s.disconnect(func)

    """
    _event = Event

    def connect(self, listener):
        self.on_trait_change(listener, '_event')

    def disconnect(self, listener):
        self.on_trait_change(listener, '_event', remove=True)

    def emit(self, value):
        self._event = value

def do_callback(dispatch, callback, *args):
    """Invoke the callback with a suitable dispatch.
    """
    if dispatch == 'ui':
        from ..utils.gui import invoke_later
        invoke_later(callback, *args)
    else:
        callback(*args)


################################################################################
# `Promise` class.
################################################################################

class Promise(HasTraits):
    """ The promise of a deferred operation, which can be used
    to add success, failure and progress callbacks for the operation.

    The Promise instance has a ``dispatch`` trait which can be set to "same" or
    "ui" and if it is set to "ui" all changes and callbacks are made in the GUI
    thread.

    """

    def __init__(self, **traits):
        HasTraits.__init__(self, **traits)
        # This should really be a reader-writer lock for performance, but this
        # will do for the time being.
        self._lock = RLock()


    # `Promise` Interface #####################################################
    def on_done(self, callback):
        """ Add callback for successful completion of the operation.
        """
        with self._lock:
            status = self._status
            if status == "pending":
                self.on_trait_change(lambda value: (self._status == 'done')
                                                    and callback(self._result),
                                    '_status', dispatch=self.dispatch)

        # Release the lock before calling the callback. Status is done so nothing
        # will mutate further - we are safe.
        if status == "done":
            do_callback(self.dispatch, callback, self._result)

    def on_error(self, callback):
        """ Add callback for failure of the operation.
        """
        with self._lock:
            status = self._status
            if status == "pending":
                self.on_trait_change(lambda value: (self._status == 'error')
                                                    and callback(self._error),
                                    '_status', dispatch=self.dispatch)

        # Release the lock before calling the callback. Status is done so nothing
        # will mutate further - we are safe.
        if status == "error":
            do_callback(self.dispatch, callback, self._error)

    def on_progress(self, callback):
        """ Add callback for progress of the operation.
        """
        with self._lock:
            status = self._status
            if status == "pending":
                self.on_trait_change(lambda value: callback(value),
                                     '_progress', dispatch=self.dispatch)

        # Release the lock before calling the callback. Status is done so nothing
        # will mutate further - we are safe.
        if self._status == 'done':
            # If operation is already completed, call with 1.0
            do_callback(self.dispatch, callback, self._progress)

    #################################
    # Traits.

    # Dispatch all callbacks either in the same thread or in the UI thread.
    dispatch = Enum('same', 'ui')

    # Status of the Promise.
    status = Property
    _status = Enum('pending', 'done', 'error')

    def _get_status(self):
        with self._lock:
            return self._status

    # The result, if any.
    result = Property
    _result = Any(Undefined)

    def _get_result(self):
        with self._lock:
            if self._status == 'pending':
                raise ValueError('Promise not completed yet')
            else:
                return self._result

    # The error, if any.
    error = Property
    _error = Any(Undefined)

    def _get_error(self):
        with self._lock:
            if self._status == 'pending':
                raise ValueError('Promise not completed yet')
            else:
                return self._error

    # The progress.
    progress = Property
    _progress = Range(0.0, 1.0, 0.0)

    def _get_progress(self):
        with self._lock:
            return self._progress



################################################################################
# `Deferred` class.
################################################################################
class Deferred(HasTraits):
    """ A Deferred operations which will complete in the future.

    Usage:
    ------
    In an asynchronous method which would normally require you to accept a
    callback method, you can use the Deferred object to let callers add
    callbacks even after making the call, in fact even after the operation
    has finished.

    The following example illustrates the simplest use of a Future::

        def background_job():
            deferred = Deferred()
            # Some async job which takes a callback
            async_background_job(callback=lambda result: deferred.done(result))
            return deferred.promise()

        def on_job_finished(self, result):
            print result

        bg_job = background_job()
        # bg_job promise can be used add callbacks
        bg_job.on_done(on_job_finished)

    The advantages of this approach to passing callbacks are:
        - ability to add callbacks after the call has been made
        - ability to add callbacks even after the function has returned
        - pass around promise objects to let others add callacks for the job
          This can be useful for example to add a progress feature to
          an existing function which would have otherwise needed to add
          an extra progress_callback argument.

    The Deferred supports two dispatch mechanisms, "same" and "ui" if the
    ``dispatch`` trait is set to "ui" all attributes are set on the GUI thread
    and all callbacks are also called from the UI thread.

    Notes:
    ------
    Always return the promise() object for an operation to callers when using
    Deferred so that so that they can only add callbacks and not set result.

    """
    # `Deferred` Interface ####################################################
    def done(self, value):
        """ Complete the deferred with success and specified result.
            and set the progress to 1.0
        """
        if self.dispatch == 'ui':
            promise = self.promise
            set_trait_later(promise, '_result', value)
            set_trait_later(promise, '_progress', 1.0)
            set_trait_later(promise, '_status', 'done')
        else:
            with self.promise._lock:
                self.promise._result = value
                self.promise._progress = 1.0
                self.promise._status = 'done'

    def error(self, value):
        """ Complete the deferred with failure and specified result. """
        if self.dispatch == 'ui':
            promise = self.promise
            set_trait_later(promise, '_error', value)
            set_trait_later(promise, '_status', 'error')
        else:
            with self.promise._lock:
                self.promise._error = value
                self.promise._status = 'error'

    def progress(self, value):
        """ Set the progress of the operation (0 <= value <= 1). """
        if self.dispatch == 'ui':
            set_trait_later(self.promise, '_progress', value)
        else:
            with self.promise._lock:
                self.promise._progress = value

    # Traits ##################################################################

    # Dispatch all callbacks either in the same thread or in the UI thread.
    dispatch = Enum('same', 'ui')

    promise = ReadOnly(Instance(Promise))

    def _promise_default(self):
        return Promise(dispatch=self.dispatch)


################################################################################
# `Future` class.
################################################################################
class Future(Promise):

    """
    This could be used for any long-running call that needs to run on
    another thread.  The result of the call is stored in the `result` trait.
    Accessing the `result` trait will block till the thread completes its task.
    The `done` method can be used to check if the thread has completed.
    Optional progress information is available in a `progress` trait.
    Additional optional information is also available in the `info` trait.

    One can also set the ``dispatch`` mode to "ui" in which case all
    changes to attributes and callbacks are made on the GUI thread.

    The following example illustrates the simplest use of a Future::

        >>> import time
        >>> x = Future(lambda : time.sleep(1) or 'done')
        >>> assert x.done() is False
        >>> assert x.result == 'done' # This will block.

    Here is a slightly more complex example illustrating the use of the
    ``f_on_progress`` and ``f_on_status`` callbacks::

        >>> def on_progress(future):
        ...    print future.progress
        ...
        >>> def on_status(future):
        ...    if future.status == 'done':
        ...        print future.result
        ...    elif future.status == 'error':
        ...        print future.error
        ...
        >>> def compute(future=None):
        ...     for i in range(5):
        ...         if future is not None:
        ...             future.info = str(i)
        ...             future.progress = (i+1)*0.2
        ...         time.sleep(0.1)
        ...     return 'done'
        ...
        >>> x = Future(compute, f_on_status=on_status,
        ...            f_on_progress=on_progress, future_kw='future')
        >>> for i in range(5):
        ...     print x.progress, x.info
        ...     time.sleep(0.1)
        ...
        >>> print x.result

    """

    # Status of the Future.
    status = DelegatesTo('promise', 'status')

    # The result of the call. When accessed this will block unless the
    # thread has finished execution.
    result = Property(Any, depends_on='status')

    # The exception if any.  ``sys.exc_info`` is stored here.
    error = DelegatesTo('promise', 'error')

    # Progress information.
    progress = Property(depends_on='promise._progress')

    # Optional information.
    info = Str('')

    #################################
    # Private Traits.

    # The thread in which the function is running.
    _thread = Instance(Thread)

    # The Deferred object for the operation and its promise.
    _deferred = Instance(Deferred)
    promise = DelegatesTo('_deferred', 'promise')

    ############################################################################
    # `object` interface.
    ############################################################################
    def __init__(self, func, f_on_status=None, f_on_progress=None,
                 future_kw=None, dispatch='same', args=None, kw=None):
        """Constructor for a Future.

        If an exception is raised when the future runs, ``sys.exc_info()``
        is stored in the error trait and the ``status`` is set to "error"

        Parameters
        ----------

        func : callable
            The callable to execute in another thread.

        f_on_status : callable
            The callable to callback when the status is changed.  This function
            will be passed the `Future` instance as a single argument.

        f_on_progress : callable
            The callable to callback when the progress is changed. This function
            will be passed the `Future` instance as a single argument.

        future_kw : str
            The keyword argument of the callable that accepts a `Future`
            instance, typically used for progress info.

        args : additional args
            These are passed on to the callable.

        kw : additional keyword args
            Passed to the callable, ``func``.

        """
        # Set this first.
        self.dispatch = dispatch
        super(Future, self).__init__()

        if f_on_progress is not None:
            self.on_progress(lambda value:f_on_progress(self))

        if f_on_status:
            self.on_done(lambda value:f_on_status(self))
            self.on_error(lambda value:f_on_status(self))

        # The wrapper function to call in a thread.
        def _f(self, *args, **kw):
            """This function is called by the `Thread` instance."""
            try:
                self._deferred.done(func(*args, **kw))
            except:
                self._deferred.error(sys.exc_info())

        # Pass self to the function if it needs it.
        if future_kw is not None and type(future_kw) in (str, utext):
            kw[future_kw] = self
        args = args or ()
        kw = kw or {}
        t = Thread(target=_f, args=(self,) + args, kwargs=kw)
        self._thread = t
        t.daemon = True
        t.start()

    ############################################################################
    # `Future` interface.
    ############################################################################
    def done(self):
        """Return True if the future has completed execution."""
        return self.promise.status != 'pending'

    ############################################################################
    # `Promise` interface.
    ############################################################################
    def on_done(self, callback):
        self.promise.on_done(callback)

    def on_error(self, callback):
        self.promise.on_error(callback)

    def on_progress(self, callback):
        self.promise.on_progress(callback)

    ############################################################################
    # Trait handlers.
    ############################################################################
    def _get_result(self):
        # _status is synchronized, so copy local to avoid constant lock acquisitions
        status = self.promise._status
        if status == 'pending':
            self._thread.join()
            # Status will have switched to "done" or "error"
            status = self.promise._status

        if status == 'done':
            return self.promise.result
        else:
            return Undefined

    def _get_progress(self):
        return self.promise.progress

    def _set_progress(self, val):
        self._deferred.progress(val)

    def __deferred_default(self):
        return Deferred(dispatch=self.dispatch)

################################################################################
# `threaded` decorator.
################################################################################
def threaded(func=None, f_on_status=None, f_on_progress=None, future_kw=None,
             dispatch='same'):
    """ A decorator to run a function in a separate thread and return a
    `Future` object which will store the results when the function completes.

    Parameters
    ----------

    future_kw : str
        The keyword argument of the decorated function that accepts a `Future`
        instance.

    f_on_status : callable
        The callable to callback when the status is changed.  This function
        will be passed the `Future` instance as a single argument.

    f_on_progress : callable
        The traits handler to callback when the progress is changed. Note
        that this is a traits handler.

    dispatch : str
        The dispatch mechanism to use.  One of either 'same' or 'ui'.

    Examples
    ---------

    The following examples illustrates its usage::

        >>> import time
        >>> @threaded
        ... def compute(x):
        ...     # long running computation
        ...     time.sleep(x)
        ...     return x
        >>> a = f(0.5)
        >>> b = f(1.0)
        >>> print a.result, b.result

        >>> @threaded(future_kw='future')
        ... def compute_with_progress(dt, future):
        ...     for i in range(10):
        ...         time.sleep(dt)
        ...         future.progress = (i+1)*0.1
        ...     return time.time()
        ...
        >>> a = compute_with_progress(0.1)
        >>> # Future is automatically passed.
    """

    def future_decorator(func, f_on_status=f_on_status,
                         f_on_progress=f_on_progress,
                         future_kw=future_kw, dispatch=dispatch):
        def _wrapper(*args, **kw):
            """The wrapper function."""
            return Future(func, f_on_status, f_on_progress, future_kw,
                          dispatch, *args, **kw)
        return wraps(func)(_wrapper)

    if func is None:
        return future_decorator
    else:
        return future_decorator(func, f_on_status, f_on_progress,
                                future_kw=future_kw, dispatch=dispatch)



#
# Jigna product code
#
# (C) Copyright 2013 Enthought, Inc., Austin, TX
# All right reserved.
#

# Standard library imports.
import logging
from os.path import abspath, dirname, join

# Local imports.
from jigna.core.interoperation import create_js_object_wrapper
from jigna.core.network_access import ProxyAccessManager
from jigna.qt import QtCore, QtGui, QtWebKit

logger = logging.getLogger(__name__)


class ProxyQWebView(QtWebKit.QWebView):

    DISABLED_ACTIONS = [
        QtWebKit.QWebPage.OpenLinkInNewWindow,
        QtWebKit.QWebPage.DownloadLinkToDisk,
        QtWebKit.QWebPage.OpenImageInNewWindow,
        QtWebKit.QWebPage.OpenFrameInNewWindow,
        QtWebKit.QWebPage.DownloadImageToDisk,
        QtWebKit.QWebPage.Reload,
        QtWebKit.QWebPage.Back
    ]

    def __init__(
        self, parent=None, python_namespace=None, callbacks=[],
        debug=True, hosts={}
    ):
        super(ProxyQWebView, self).__init__(parent)

        self._page = ProxyQWebPage()
        self.setPage(self._page)

        # Connect JS with python.
        self.expose_python_namespace(python_namespace, callbacks)

        # Install custom access manager to delegate requests to custom WSGI
        # hosts.
        self._access_manager = ProxyAccessManager(hosts=hosts)
        self._page.setNetworkAccessManager(self._access_manager)

        # Disable some actions
        for action in self.DISABLED_ACTIONS:
            self.pageAction(action).setVisible(False)

        # Setup debug flag
        self._page.settings().setAttribute(
            QtWebKit.QWebSettings.DeveloperExtrasEnabled, debug
        )

        # Set sizing policy
        self.setSizePolicy(
            QtGui.QSizePolicy.Expanding, QtGui.QSizePolicy.Expanding
        )

    def execute_js(self, js):
        """ Execute JavaScript synchronously.

        Warning: under most circumstances, this method should not be called
        when the page is loading.

        """
        frame = self._page.mainFrame()
        result = frame.evaluateJavaScript(js)
        result = self._apply_null_fix(result)

        return result

    def expose_python_namespace(self, python_namespace, callbacks):
        """ Exposes the given python namespace to Javascript.

        Javascript can access the given list of callbacks as if they were
        methods on the object described by the python namespace.

        python_namespace: str:
            Namespace to expose to the JS world. This creates an object of the
            same name and attaches it to window frame.

        callbacks: [method_name: callable]:
            This list of callbacks is what is exposed to the JS world via the
            given python namespace.

        Usage:
        ------

        For example, doing this::

            expose_python_namespace('python', ['say_hello', say_hello])

        will create a window level object on the JS side which looks like
        this::

            window.python.say_hello == <a function which calls Python land>

        """
        frame = self._page.mainFrame()
        js_wrapper = create_js_object_wrapper(callbacks=callbacks,parent=frame)
        frame.javaScriptWindowObjectCleared.connect(
            lambda: self._on_js_window_cleared(python_namespace, js_wrapper)
        )

    def _on_js_window_cleared(self, namespace, js_wrapper):
        frame = self._page.mainFrame()
        frame.addToJavaScriptWindowObject(namespace, js_wrapper)

    def setUrl(self, url):
        """ Reimplemented to make sure that when we return, the DOM is ready to
        use.

        Based on the local event loop approach described here:
        http://doc.qt.digia.com/qq/qq27-responsive-guis.html#waitinginalocaleventloop
        """
        event_loop = QtCore.QEventLoop()
        self._loaded = False

        def on_load():
            self._loaded = True
            event_loop.quit()

        self._page.loadFinished.connect(on_load)

        super(ProxyQWebView, self).setUrl(QtCore.QUrl(url))

        if not self._loaded:
            event_loop.exec_()

    #### Private protocol #####################################################

    @staticmethod
    def _apply_null_fix(obj):
        """ Makes sure that None objects coming from Qt bridge are actually
        None.

        We need this because NoneType objects coming from PyQt are of a
        `QPyNullVariant` type, not None. This method converts such objects to
        the standard None type.

        """
        if isinstance(obj, getattr(QtCore, 'QPyNullVariant', type(None))):
            return None

        return obj


class ProxyQWebPage(QtWebKit.QWebPage):
    """ Overridden to open external links in a web browser.

    Source: http://www.expobrain.net/2012/03/01/open-urls-in-external-browser-by-javascript-in-webkit/
    """

    def acceptNavigationRequest(self, frame, request, type):
        # Checking this is same as checking if the client side <a> tag making
        # the HTTP request had target="_blank" as an attribute.
        if frame is None:
            import webbrowser
            webbrowser.open_new(request.url().toString())

            return False

        else:
            return super(ProxyQWebPage, self).acceptNavigationRequest(
                frame, request, type
            )

    def createWindow(self, *args, **kwargs):
        return ProxyQWebPage()

if __name__ == '__main__':
    app = QtGui.QApplication([])
    w = ProxyQWebView()
    w.show()
    w.raise_()
    w.load(QtCore.QUrl('http://www.google.com/'))
    app.exec_()



from __future__ import absolute_import
import sys
from textwrap import dedent
from threading import Thread
import time
import sys
import unittest
from unittest import skip

try:
    from tornado.ioloop import IOLoop
except ImportError:
    raise unittest.SkipTest("Tornado not installed")

try:
    from selenium import webdriver
except ImportError:
    raise unittest.SkipTest("selenium not installed")

# Local imports.
from jigna.utils.web import get_free_port
from .test_jigna_qt import (TestJignaQt, Person, body_html, AddressBook,
    sleep_while)


def patch_sys_modules():
    backup = sys.modules.copy()
    for name in list(sys.modules.keys()):
        if 'PySide' in name or 'PyQt' in name:
            del sys.modules[name]

    return backup


def restore_sys_modules(backup):
    sys.modules = backup


def assert_no_qt_in_sys_modules():
    for name in list(sys.modules.keys()):
        if 'PySide' in name or 'PyQt' in name:
            raise AssertionError("Qt import found: %s." % name)


def start_io_loop():
    ioloop = IOLoop.instance()
    ioloop.make_current()
    ioloop.start()


def stop_io_loop():
    IOLoop.instance().stop()


class TestJignaWebSync(TestJignaQt):
    @classmethod
    def setUpClass(cls, async=False):

        cls._backup_modules = patch_sys_modules()

        from jigna.template import Template
        from jigna.web_app import WebApp
        fred = Person(name='Fred', age=42)
        addressbook = AddressBook()
        template = Template(body_html=body_html, async=async)
        port = get_free_port()
        app = WebApp(
            template=template,
            context={'model':fred, 'addressbook': addressbook}, async=async
        )
        app.listen(port)

        # Start the tornado server in a different thread so that we can write
        # test statements here in the main loop.
        t = Thread(target=start_io_loop)
        t.setDaemon(True)
        t.start()

        browser = webdriver.Firefox()
        browser.implicitly_wait(10)
        browser.get('http://localhost:%d' % port)
        cls.app = app
        cls.fred = fred
        cls.browser = browser
        cls.addressbook = addressbook
        cls.thread = t

    @classmethod
    def tearDownClass(cls):
        cls.browser.quit()

        IOLoop.instance().add_callback(stop_io_loop)
        thread = cls.thread
        count = 0
        while thread.is_alive() and count < 50:
            time.sleep(0.1)
        cls.thread.join()

        # Smells a bit but doing this here ensures that none of the tested
        # cases imports Qt.
        assert_no_qt_in_sys_modules()

        restore_sys_modules(cls._backup_modules)

    def setUp(self):
        cls = self.__class__
        self.app = cls.app
        self.browser = cls.browser
        self.fred = cls.fred
        self.fred.spouse = None
        self.fred.fruits = []
        self.fred.friends = []
        # Wait for the model to be setup before running the tests.
        self.get_attribute('jigna.models.model.name', self.fred.name)

    def process_events(self):
        pass

    def execute_js(self, js):
        return self.browser.execute_script(js)

    def get_attribute(self, js, expect):
        get_js = dedent("""
        var result;
        try {
            result = eval(%r);
        } catch (err) {
            result = undefined;
        }
        return result;
        """%js)
        self.execute_js(get_js)

        check_js = get_js
        result = self.execute_js(check_js)
        count = 0
        while result != expect and count < 20:
            time.sleep(0.1)
            result = self.execute_js(check_js)
            count += 1
        return result

    def assertJSEqual(self, js, value):
        result = self.get_attribute(js, value)
        if isinstance(value, (list, tuple)):
            msg = "Lengths different: expected %d, got %d" % \
                (len(value), len(result))
            self.assertEqual(len(value), len(result), msg)
            for index in range(len(value)):
                expect = value[index]
                got = result[str(index)]
                if got != expect:
                    got = self.get_attribute(js+"['%d']"%index, expect)
                msg = "%s[%s] != %s, got %s"%(js, index, expect, got)
                self.assertEqual(expect, got, msg)
        else:
            msg = "%s != %s, got %s"%(js, value, result)
            self.assertEqual(value, result, msg)

    def test_instance_trait(self):
        # Overridden to work with the web backend.
        self.assertJSEqual("jigna.models.model.spouse", None)
        wilma = Person(name='Wilma', age=40)
        self.fred.spouse = wilma
        self.assertJSEqual("jigna.models.model.spouse.name", 'Wilma')
        self.assertJSEqual("jigna.models.model.spouse.age", 40)

        # Set in the JS side.
        self.execute_js("jigna.models.model.spouse.name = 'Wilmaji'")
        self.execute_js("jigna.models.model.spouse.age = 41")
        self.wait_and_assert(lambda: wilma.name != 'Wilmaji')
        self.wait_and_assert(lambda: wilma.age != 41)

    def test_reload_works_correctly(self):
        # Given
        fred = self.fred
        fred.fruits = ["peach", "pear"]
        dino = Person(name="Dino", age=10)
        fred.friends = [dino]
        fred.phonebook = {'joe' : 123, 'joan' : 345}
        wilma = Person(name='Wilma', age=40)
        fred.spouse = wilma

        # When
        self.execute_js("window.location.reload();")
        time.sleep(1)
        self.get_attribute("jigna.models.model.name", fred.name)

        # Then
        self.assertJSEqual("jigna.models.model.name", fred.name)
        self.assertJSEqual("jigna.models.model.friends[0].name", "Dino")
        self.assertJSEqual("jigna.models.model.friends[0].age", 10)
        self.assertJSEqual("jigna.models.model.phonebook.joe", 123)
        self.assertJSEqual("jigna.models.model.phonebook.joan", 345)
        self.assertJSEqual("jigna.models.model.phonebook", fred.phonebook)
        self.assertJSEqual("jigna.models.model.fruits", fred.fruits)
        self.assertJSEqual("jigna.models.model.spouse.name", 'Wilma')
        self.assertJSEqual("jigna.models.model.spouse.age", 40)

# Delete this so running just this file does not run all the tests.
del TestJignaQt

if __name__ == "__main__":
    unittest.main()

#### EOF ######################################################################



# Standard library imports
import threading
import mimetypes
import logging
from os.path import exists, join, sep

# Enthought library imports
from traits.api import HasTraits, Str, Dict, Directory, on_trait_change

mimeLock = threading.Lock()
mimeInitialized = False

logger = logging.getLogger(__name__)


def guess_type(path):
    """ Thread-safe wrapper around the @#%^$$# non-thread safe mimetypes module. NEVER
    call mimetypes directly.

    """

    global mimeLock
    global mimeInitialized

    if not mimeInitialized:
        with mimeLock:
            if not mimeInitialized:
                mimetypes.init()

                # On Windows, the mime type for 'ttf' (True Type Fonts) isn't
                # recognized correctly. We, therefore, add the correct mime
                # type here.
                mimetypes.add_type('application/font-sfnt', '.ttf')

                mimeInitialized = True

    guessed = mimetypes.guess_type(path)

    return (guessed[0] or "", guessed[1] or "")


class FileLoader(HasTraits):

    #: Root directory where it looks
    root = Directory

    #: A dictionary of overrides which holds canned responses for some special
    #: paths
    overrides = Dict

    #### WSGI protocol ########################################################

    def __call__(self, env, start_response):

        # Clean up path and handle Windows paths
        path = env['PATH_INFO'].strip('/').replace('/', sep)

        # Check if it is handled by one of the overrides
        if self.overrides.get(path) is not None:
            start_response(
                '200 OK', [('Content-Type', '; '.join(guess_type(path)))]
            )
            return [self.overrides[path]]

        # Continue, if the path wasn't handled by canned responses for special
        # paths
        path = join(self.root, path)
        if not exists(path):
            start_response('404 File not found', [])
            return [""]

        else:
            start_response(
                '200 OK', [('Content-Type', '; '.join(guess_type(path)))]
            )
            with open(path, 'rb') as f:
                response = f.read()
            return [response]



"""
This example shows two-way data binding on an `Instance` trait.
"""

#### Imports ####
from __future__ import print_function

from traits.api import HasTraits, Instance, Str
from jigna.api import HTMLWidget, Template
from jigna.qt import QtGui
from jigna.utils.gui import do_after

#### Domain model ####

class Person(HasTraits):
    name = Str
    spouse = Instance('Person')

    def marry(self, person):
        self.spouse = person
        person.spouse = self

#### UI layer ####

body_html = """
    <div>
      Name: <input ng-model="person.name"><br/>
      Spouse's name: <input ng-model="person.spouse.name"><br/>
      Spouse's spouse's name: <input ng-model="person.spouse.spouse.name">
    </div>
"""

template = Template(body_html=body_html)


#### Entry point ####

def main():
    # Start the Qt application
    app = QtGui.QApplication([])

    # Instantiate the domain model
    fred = Person(name='Fred')
    wilma = Person(name='Wilma')

    # Create the jigna based HTML widget which renders the given HTML template
    # with the given context.
    widget = HTMLWidget(template=template, context={'person':fred})
    widget.show()

    # Schedule some operations on the domain model.
    #
    # The operation should fill in the `spouse` field of the person and this
    # should be reflected in the UI.
    do_after(2500, fred.marry, wilma)

    # Start the event loop.
    #
    # Initially, the `spouse` field of the person is empty, so the fields
    # related to the spouse should be empty in the UI.
    app.exec_()

    # Check the final values of the instance
    print(fred.name, fred.spouse.name, wilma.name)

if __name__ == "__main__":
    main()

#### EOF ######################################################################



#
# (C) Copyright 2013-2016 Enthought, Inc., Austin, TX
# All right reserved.
#

""" Web implementations of the Jigna Server and Bridge. """


# Standard library.
import json
import mimetypes
from os.path import abspath, dirname, join
import threading
import traceback
try:
    from urllib import unquote
except ImportError:
    # The above import will not work on Python-3.x.
    from urllib.parse import unquote

# 3rd party library.
from tornado.websocket import WebSocketHandler
from tornado.web import Application, RequestHandler, StaticFileHandler
from tornado.ioloop import IOLoop

# Enthought library.
from traits.api import (
    Bool, List, Str, Instance, TraitDictEvent, TraitListEvent
)

# Jigna library.
from jigna.server import Bridge, Server
from jigna.core.wsgi import guess_type

#: Path to jigna.js file
JIGNA_JS_FILE = join(abspath(dirname(__file__)), 'js', 'dist', 'jigna.js')


def normalize_slice(s, size):
    """ Normalize a python slice such that.

    s - slice instance to normalize
    size - the size of original sequence to normalize for

    The returned slice instance has following properties:
      - start, stop, step are all int
      - start < stop
      - step > 1
    """
    start, stop, step = s.indices(size)
    if step < 0:
        step = -step
        start, stop = stop, start
        diff = stop - start
        if diff % step == 0:
            n = diff//step - 1
        else:
            n = diff//step
        start = stop - n*step

        stop += 1

    return slice(start, stop, step)


class WebBridge(Bridge):
    """ Bridge that handles the client-server communication. """

    #### 'Bridge' protocol ####################################################

    def send_event(self, event):
        """ Send an event. """

        # Tornado does not support multiple threads calling send_message.
        # Instead one should add a callback on the IOLoop instance as done
        # below.  See:
        # http://www.tornadoweb.org/en/stable/web.html?highlight=thread#thread-safety-notes

        main_thread = isinstance(
            threading.current_thread(), threading._MainThread
        )

        try:
            jsonized_event = json.dumps(event)
        except TypeError:
            return

        message_id = -1
        data = json.dumps([message_id, jsonized_event])
        for socket in self._active_sockets:
            if main_thread:
                socket.write_message(data)
            else:
                IOLoop.instance().add_callback(socket.write_message, data)

        return

    #### 'WebBridge' protocol #################################################

    def add_socket(self, socket):
        """ Add a client socket. """

        self._active_sockets.append(socket)

        return

    def remove_socket(self, socket):
        """ Remove a client socket. """

        self._active_sockets.remove(socket)

        return

    #### Private protocol #####################################################

    #: All active client sockets.
    _active_sockets = List


class WebServer(Server):
    """ Web-based server implementation.

    This implementation uses web-sockets for making two-way communication.

    """

    ### 'WebServer' protocol ##################################################

    #: The list of tornado handlers which map URL patterns to callables
    handlers = List
    def _handlers_default(self):
        return [
            # To serve jigna.js from package source
            (
                r"/jigna/(.*)", StaticFileHandler,
                dict(path=dirname(JIGNA_JS_FILE))
            ),

            # This handler handles the web socket based async version of the
            # web interface.
            (
                r"/_jigna_ws", AsyncWebSocketHandler,
                dict(bridge=self._bridge, server=self)
            ),

            # This handler handles synchronous GET requests from JS proxy
            # getters and returns the corresponding attribute from the
            # python side.
            (r"/_jigna", SyncGETHandler, dict(server=self)),

            # Main handler which returns the jigna HTML and other resources
            # by resolving it via server's base url.
            (r".*", MainHandler, dict(server=self)),
        ]

    #: The trait change dispatch mechanism to use when traits change.
    trait_change_dispatch = Str('same')

    #### Private protocol #####################################################

    _bridge = Instance(WebBridge)
    def __bridge_default(self):
        return WebBridge()


class AsyncWebServer(WebServer):
    """ Asynchronous Web-based server implementation.

    This implementation uses web-sockets for making two-way communication.
    A GET request is not used in this case.  This allows one to embed
    jigna on another web application.

    """

    def _get_attribute_values(self, obj, attribute_names):
        """ Get the values of all 'public' attributes on an object.

        Return a list of strings.

        """

        return [self._marshal(self._get_attribute_default(obj, name))
                for name in attribute_names]

    def _get_attribute_default(self, obj, name):
        value = getattr(obj, name, None)
        if isinstance(value, list):
            value = []
        elif isinstance(value, dict):
            value = {}
        elif hasattr(value, '__dict__'):
            pass
        elif value is None:
            pass
        else:
            value = type(value)()
        return value

    def _get_dict_info(self, obj):
        """ Get a description of a dict. """
        values = self._get_list_info(list(obj.values()))
        return dict(keys=list(obj.keys()), values=values)

    def _get_instance_info(self, obj):
        """ Get a description of an instance. """

        info = super(WebServer, self)._get_instance_info(obj)

        # If this is a new type, also send the attribute_values.
        if 'attribute_names' in info:
            attribute_values = self._get_attribute_values(
                obj, info['attribute_names']
            )
            info['attribute_values'] = attribute_values
            self._send_new_type_event(info)
            # Now that the type info is sent we do not need to send all that
            # information again.
            info = dict(type_name=info['type_name'])

        return info

    def _get_list_info(self, obj):
        """ Get a description of a list. """
        data = self._marshal_all(obj)
        return dict(length=len(obj), data=data)

    def _send_object_changed_event(self, obj, trait_name, old, new):
        """ Send an object changed event. """

        if trait_name.startswith('_'):
            return

        if isinstance(new, TraitListEvent):
            trait_name  = trait_name[:-len('_items')]
            trait = getattr(obj, trait_name)
            value = id(trait)
            if isinstance(new.index, slice):
                # Handle an extended slice.  Note that one cannot increase the
                # size of the list here.  So one is either deleting elements
                # or changing them.
                added = new.added[0] if len(new.added) > 0 else []
                removed = new.removed[0]
                prev_size = len(trait) + len(removed) - len(added)
                s = normalize_slice(new.index, prev_size)
                if new.index.step < 0:
                    # When the step is negative we swap the order of start and
                    # stop so that start <= stop, we therefore must reverse the
                    # added elements.
                    added = added[::-1]
                info = dict(
                    start=s.start, stop=s.stop, step=s.step,
                    removed=len(removed),
                    added=self._get_list_info(added)
                )
            else:
                # This information can be used by the Array.splice method.
                info = dict(
                    index=new.index, removed=len(new.removed),
                    added=self._get_list_info(new.added)
                )

            data = dict(type='list', value=value, info=info)
            items_event = True

        elif isinstance(new, TraitDictEvent):
            trait_name  = trait_name[:-len('_items')]
            trait = getattr(obj, trait_name)
            value = id(trait)
            added, removed = new.added, list(new.removed.keys())
            for key in new.changed:
                added[key] = trait[key]
            info = dict(removed=removed, added=self._get_dict_info(added))
            data = dict(type='dict', value=value, info=info)
            items_event = True

        else:
            # fixme: intent is non-scalar or maybe container?
            if hasattr(new, '__dict__') or isinstance(new, (dict, list)):
                self._register_object(new)

            data = self._marshal(new)
            items_event = False

        event = dict(
            obj  = str(id(obj)),
            name = trait_name,
            # fixme: This smells a bit, but marshalling the new value gives us
            # a type/value pair which we need on the client side to determine
            # what (if any) proxy we need to create.
            data = data,

            # fixme: This is how we currently detect an 'xxx_items' event on
            # the JS side.
            items_event = items_event
        )

        self.send_event(event)

        return

    def _send_new_type_event(self, data):
        """Send a new_type event.  The data passed is the type information
        dict.
        """
        event = dict(
            obj  = 'jigna',
            name = 'new_type',
            data = data
        )
        self.send_event(event)


##### Request handlers ########################################################


class MainHandler(RequestHandler):

    def initialize(self, server):
        self.server = server

        return

    def get(self):
        path = unquote(self.request.path[1:])
        if not len(path):
            self.write(self.server.html)
        else:
            mime_type, _ = guess_type(path)
            self.set_header('Content-Type', mime_type)
            self.write(open(join(self.server.base_url, path), 'rb').read())

        return


class SyncGETHandler(RequestHandler):

    def data_received(self, chunk):
        pass

    def initialize(self, server):
        self.server = server
        return

    def get(self):
        jsonized_request = self.get_argument("data")

        jsonized_response = self.server.handle_request(jsonized_request)
        self.write(jsonized_response)
        return


class AsyncWebSocketHandler(WebSocketHandler):

    def initialize(self, bridge, server):
        self.bridge = bridge
        self.server = server
        return

    def open(self):
        self.bridge.add_socket(self)
        return

    def on_message(self, message):
        try:
            request_id, jsonized_request = json.loads(message)
            jsonized_response = self.server.handle_request(jsonized_request)
            self.write_message(json.dumps([request_id, jsonized_response]))
        except Exception:
            traceback.print_exc()
            self.write_message(json.dumps([request_id, '{}']))
        return

    def on_close(self):
        self.bridge.remove_socket(self)
        return

    def write_message(self, msg, binary=False):
        return super(AsyncWebSocketHandler, self).write_message(msg, binary)

#### EOF ######################################################################



"""
This example shows how to initialize Jigna's HTML interface by reading
a full html file, rather than specifying body_html and head_html.
"""

#### Imports ####
from __future__ import print_function

from traits.api import HasTraits, Int, Str
from jigna.api import HTMLWidget, Template
from jigna.qt import QtGui

#### Domain model ####

class Person(HasTraits):
    name = Str
    age  = Int

    def update_name(self, name):
        self.name = name

#### UI layer ####

template = Template(html_file='ex2_simple_view_full.html')

#### Entry point ####

def main():
    # Start the Qt application
    app = QtGui.QApplication([])

    # Instantiate the domain model
    fred = Person(name='Fred', age=42)

    # Create the jigna based HTML widget which renders the given HTML template
    # with the given context.
    widget = HTMLWidget(template=template, context={'person': fred})
    widget.show()

    # Start the event loop
    app.exec_()

    # Check the values after the UI is closed
    print(fred.name, fred.age)

if __name__ == "__main__":
    main()

#### EOF ######################################################################



import os
import sys
import tempfile
import unittest

import mock

from tornado.web import Application
from tornado.httputil import HTTPServerRequest

from jigna.web_server import MainHandler, normalize_slice

# A dummy image to write and test with.
DATA = b"""\x89PNG\r\n\x1a\n\x00\x00\x00\rIHDR\x00\x00\x00\x05\x00\x00\x00\x05\x08\x06\x00\x00\x00\x8do&\xe5\x00\x00\x00\x04gAMA\x00\x00\xb1\x8f\x0b\xfca\x05\x00\x00\x00 cHRM\x00\x00z&\x00\x00\x80\x84\x00\x00\xfa\x00\x00\x00\x80\xe8\x00\x00u0\x00\x00\xea`\x00\x00:\x98\x00\x00\x17p\x9c\xbaQ<\x00\x00\x00\tpHYs\x00\x00\x0b\x13\x00\x00\x0b\x13\x01\x00\x9a\x9c\x18\x00\x00\x01YiTXtXML:com.adobe.xmp\x00\x00\x00\x00\x00<x:xmpmeta xmlns:x="adobe:ns:meta/" x:xmptk="XMP Core 5.4.0">\n   <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#">\n      <rdf:Description rdf:about=""\n            xmlns:tiff="http://ns.adobe.com/tiff/1.0/">\n         <tiff:Orientation>1</tiff:Orientation>\n      </rdf:Description>\n   </rdf:RDF>\n</x:xmpmeta>\nL\xc2\'Y\x00\x00\x00tIDAT\x08\x1d\x01i\x00\x96\xff\x01\x00\x1cj\xff}e0\x00;8*\x00\xcb\xcd\xd9\x00\xa2\xad\xd3\x00\x04gP!\x00<9)\x00\x03\x03\x03\x00YVC\x00\xd7\xd9\xe5\x00\x04\x08\x08\x01\x00\xb0\xb4\xc3\x00\n\x08\r\x00\x0f\x0e\x08\x00\xf7\xf8\xfd\x00\x04\xe1\xe3\xf1\x0030\x18\x00\xfc\xfb\x03\x00>>0\x00\x04\x05\x03\x00\x03\xef\xff0\x80\xef\xed\xf3\x00>:$\x00\xdc\xdc\xe5\x00y\x88\xc9\x00\x9a\xa5"\x98\x19\x929\xa9\x00\x00\x00\x00IEND\xaeB`\x82"""


class DummyServer:
    def __init__(self):
        self.html = b'html'
        self.base_url = '/'


class TestableMainHandler(MainHandler):
    def write(self, data):
        self.test_data = data


class TestMainHandler(unittest.TestCase):

    def setUp(self):
        self.tmpfile = ''
        self.fd = None

    def tearDown(self):
        if len(self.tmpfile) > 0 and os.path.exists(self.tmpfile):
            os.close(self.fd)
            os.remove(self.tmpfile)

    def _make_text_data(self):
        self.fd, self.tmpfile = tempfile.mkstemp('.txt')
        with open(self.tmpfile, 'w') as fp:
            fp.write('hello')

    def _make_binary_data(self):
        self.fd, self.tmpfile = tempfile.mkstemp('.png')
        with open(self.tmpfile, 'wb') as fp:
            fp.write(DATA)

    def _make_request(self, path):
        request = mock.MagicMock(spec=HTTPServerRequest)()
        if sys.platform.startswith('win'):
            request.path = '/' + path
        else:
            request.path = path
        return request

    def test_get_root(self):
        # Given
        request = self._make_request('')
        app = Application()
        server = DummyServer()

        # When
        h = TestableMainHandler(app, request, server=server)
        h.get()

        # Then
        self.assertEqual(h.test_data, b'html')

    def test_get_text_data(self):
        # Given
        self._make_text_data()
        request = self._make_request(self.tmpfile)
        app = Application()
        server = DummyServer()

        # When
        h = TestableMainHandler(app, request, server=server)
        h.get()

        # Then
        self.assertEqual(h.test_data, b'hello')

    def test_get_binary_data(self):
        # Given
        self._make_binary_data()
        request = self._make_request(self.tmpfile)
        app = Application()
        server = DummyServer()

        # When
        h = TestableMainHandler(app, request, server=server)
        h.get()

        # Then
        self.assertEqual(h.test_data, DATA)


class TestNormalizeSlice(unittest.TestCase):
    def test_simple_slice(self):
        # Given
        s = slice(1, 10)

        # When
        s1 = normalize_slice(s, 10)

        # Then
        self.assertEqual(s1, slice(1, 10, 1))

        # Given
        s = slice(1, 10, 2)

        # When
        s1 = normalize_slice(s, 10)

        # Then
        self.assertEqual(s1, slice(1, 10, 2))

    def test_slice_with_missing_start_stop(self):
        # Given
        s = slice(None, 10)

        # When
        s1 = normalize_slice(s, 10)

        # Then
        self.assertEqual(s1, slice(0, 10, 1))

        # Given
        s = slice(1, None)

        # When
        s1 = normalize_slice(s, 10)

        # Then
        self.assertEqual(s1, slice(1, 10, 1))

    def test_slice_with_negative_indices(self):
        # Given
        s = slice(-5, -1)

        # When
        s1 = normalize_slice(s, 10)

        # Then
        self.assertEqual(s1, slice(5, 9, 1))

    def test_slice_with_negative_stride(self):
        # Given
        s = slice(None, None, -2)

        # When
        s1 = normalize_slice(s, 10)

        # Then
        self.assertEqual(s1, slice(1, 10, 2))
        data = list(range(10))
        self.assertEqual(data[s], data[s1][::-1])

        # Given
        s = slice(None, None, -2)

        # When
        s1 = normalize_slice(s, 11)

        # Then
        self.assertEqual(s1, slice(0, 11, 2))
        data = list(range(11))
        self.assertEqual(data[s], data[s1][::-1])

        # Given
        s = slice(None, None, -3)

        # When
        s1 = normalize_slice(s, 10)

        # Then
        self.assertEqual(s1, slice(0, 10, 3))
        data = list(range(10))
        self.assertEqual(data[s], data[s1][::-1])

        # Given
        s = slice(-1, -5, -2)

        # When
        s1 = normalize_slice(s, 10)

        # Then
        self.assertEqual(s1, slice(7, 10, 2))
        data = list(range(10))
        self.assertEqual(data[s], data[s1][::-1])


if __name__ == '__main__':
    unittest.main()



"""
This example demonstrates how to schedule updates to the application model in
the web version.
"""

#### Imports ####
from __future__ import print_function

import datetime
from jigna.web_app import WebApp
from jigna.template import Template
from tornado.ioloop import IOLoop
from traits.api import HasTraits, Str

#### Domain model ####

class MOTD(HasTraits):
    message = Str

    def update_message(self, message):
        self.message = message

#### UI layer ####

body_html = """
    <div>
      Message of the day: {{motd.message}}
    </div>
"""

template = Template(body_html=body_html)

#### Entry point ####

def main():
    # Obtain the tornado ioloop object
    ioloop = IOLoop.instance()

    # Instantiate the domain model
    motd = MOTD(message="Explicit is better than implicit.")

    # Create a web app serving the view with the domain model added to its
    # context.
    app = WebApp(template=template, context={'motd':motd})
    app.listen(8000)

    # Schedule an update to a model variable after 10 seconds. If the user's
    # browser is connected to the web app before 10 seconds, it will see the
    # change in the model variable happening. Otherwise, it will only see the
    # updated value.
    #
    # NOTE: The way to schedule a task in the future has been greatly improved
    # in Tornado 4.1.dev version. You can specify additional arguments to
    # methods etc.
    delay = datetime.timedelta(0, 10)
    ioloop.add_timeout(delay,
        lambda : motd.update_message("Flat is better than nested.")
    )

    # Start serving the web app on port 8000.
    print('Serving on port 8000...')
    ioloop.start()

if __name__ == "__main__":
    main()

#### EOF ######################################################################



"""
This example shows two-way data binding for the `List` traits whose items are of
the primitive type.
"""

#### Imports ####
from __future__ import print_function

from traits.api import HasTraits, Str, List
from jigna.api import HTMLWidget, Template
from jigna.qt import QtGui
from jigna.utils.gui import do_after

#### Domain model ####

class Basket(HasTraits):
    fruits = List(Str)

#### UI layer ####

# Looping over the list of primitive variables using `ng-repeat` has two gotchas
# (both of them are standard AngularJS gotchas):
#
# 1. We need to add `track by $index` for looping through primitive variables
# otherwise angular creates new DOM nodes for every edit of the variable. One
# manifestation of this behaviour is that the input field gets out of focus for
# every edit since each edit really creates a new input field node.
# See http://www.bennadel.com/blog/2556-using-track-by-with-ngrepeat-in-angularjs-1-2.htm
#
# 2. We can't write to the `fruit` variable by making a local assignment i.e.
# `ng-model="fruit"`. This is because `fruit` refers to a local scope variable,
# so any writes to this local variable won't update the model values in Python
# which are bound to the $rootScope.
# See http://stackoverflow.com/questions/15488342/binding-inputs-to-an-array-of-primitives-using-ngrepeat-uneditable-inputs

body_html = """
    <div>
      Fruits in the basket:
      <div ng-repeat="fruit in basket.fruits track by $index">
        <input ng-model="basket.fruits[$index]">
      </div>
    </div>
"""

template = Template(body_html=body_html)

#### Entry point ####

def main():
    # Start the Qt application
    app = QtGui.QApplication([])

    # Instantiate the domain model
    basket = Basket(fruits=['peach', 'pear'])

    # Create the jigna based HTML widget which renders the given HTML template
    # with the given context.
    widget = HTMLWidget(template=template, context={'basket':basket})
    widget.show()

    # Schedule some operations on the list.
    #
    # We're trying to append and insert primitives to the list in the future.
    # This should be reflected in the UI.
    do_after(2500, basket.fruits.append, 'mango')
    do_after(5000, basket.fruits.insert, 0, 'banana')

    # Start the event loop
    app.exec_()

    # Check the final values of the list attribute
    print(basket.fruits)

if __name__ == "__main__":
    main()

#### EOF ######################################################################



from __future__ import absolute_import
import sys
from threading import Thread
import unittest

try:
    from tornado.ioloop import IOLoop
except ImportError:
    raise unittest.SkipTest("Tornado not installed")

try:
    from selenium import webdriver
except ImportError:
    raise unittest.SkipTest("selenium not installed")

# Local imports.
from jigna.utils.web import get_free_port
from .test_jigna_web import TestJignaWebSync, Person, AddressBook, \
    patch_sys_modules, start_io_loop
from .test_jigna_vue_qt import body_vue_html


class TestJignaVueWebSync(TestJignaWebSync):
    @classmethod
    def setUpClass(cls, async=False):
        cls._backup_modules = patch_sys_modules()
        from jigna.vue_template import VueTemplate
        from jigna.web_app import WebApp
        fred = Person(name='Fred', age=42)
        addressbook = AddressBook()
        template = VueTemplate(body_html=body_vue_html, async=async)
        port = get_free_port()
        app = WebApp(
            template=template,
            context={'model':fred, 'addressbook':addressbook},
            async=async
        )
        app.listen(port)

        # Start the tornado server in a different thread so that we can write
        # test statements here in the main loop
        t = Thread(target=start_io_loop)
        t.setDaemon(True)
        t.start()

        # Note: Firefox needs geckodriver with recent releases
        # https://github.com/mozilla/geckodriver/releases
        browser = webdriver.Firefox()

        browser.get('http://localhost:%d'%port)
        cls.app = app
        cls.fred = fred
        cls.browser = browser
        cls.addressbook = addressbook
        cls.thread = t


del TestJignaWebSync


if __name__ == "__main__":
    unittest.main()



# Standard library imports
import sys

# Local imports
from ..qt import QtGui, QtCore

def ui_handler(handler, *args, **kw):
    """ Handles UI notification handler requests that occur on a thread other
    than the UI thread.
    """
    invoke_later(handler, *args, **kw)

def set_trait_later(obj, trait_name, value):
    """ Set the given trait name on the given object in the GUI thread.
    """
    invoke_later(setattr, obj, trait_name, value)

def invoke_later(callable, *args, **kw):
    """ Invoke the callable in the GUI thread.
    """
    _FutureCall(0, callable, *args, **kw)

def do_after(ms, callable, *args, **kw):
    """ Invoke the callable after the given number of milliseconds.
    """
    app = QtGui.QApplication.instance() or QtGui.QApplication(sys.argv)
    QtCore.QTimer.singleShot(ms, lambda : callable(*args, **kw))

def process_events():
    """ Process all events.
    """
    QtCore.QCoreApplication.processEvents(QtCore.QEventLoop.AllEvents)

#### Private protocol #########################################################

class _FutureCall(QtCore.QObject):
    """ This is a helper class to facilitate execution of a function in the
    future (after all events are finished processing) in the GUI thread.
    """

    # Keep a list of references so that they don't get garbage collected.
    _calls = []

    # Manage access to the list of instances.
    _calls_mutex = QtCore.QMutex()

    # A new Qt event type for _FutureCalls
    _gui_event = QtCore.QEvent.Type(QtCore.QEvent.registerEventType())

    def __init__(self, ms, callable, *args, **kw):
        super(_FutureCall, self).__init__()

        # Save the arguments.
        self._ms = ms
        self._callable = callable
        self._args = args
        self._kw = kw

        # Save the instance.
        self._calls_mutex.lock()
        self._calls.append(self)
        self._calls_mutex.unlock()

        # Move to the main GUI thread.
        self.moveToThread(QtGui.QApplication.instance().thread())

        # Post an event to be dispatched on the main GUI thread. Note that
        # we do not call QTimer.singleShot here, which would be simpler, because
        # that only works on QThreads. We want regular Python threads to work.
        event = QtCore.QEvent(self._gui_event)
        QtGui.QApplication.postEvent(self, event)
        QtGui.QApplication.sendPostedEvents()

    def event(self, event):
        """ QObject event handler.
        """
        if event.type() == self._gui_event:
            # Invoke the callable (puts it at the end of the event queue)
            QtCore.QTimer.singleShot(self._ms, self._dispatch)
            return True

        return super(_FutureCall, self).event(event)

    def _dispatch(self):
        """ Invoke the callable.
        """
        try:
            self._callable(*self._args, **self._kw)
        finally:
            self._finished()

    def _finished(self):
        """ Remove the call from the list, so it can be garbage collected.
        """
        self._calls_mutex.lock()
        try:
            self._calls.remove(self)
        finally:
            self._calls_mutex.unlock()


#
# (C) Copyright 2013-2016 Enthought, Inc., Austin, TX
# All right reserved.
#
""" Abstract base classes for the Jigna server and bridge. """


# Standard library.
import inspect
import json
import logging
import traceback

# Enthought library.
from traits.api import (
    Any, Dict, Event, HasTraits, Instance, Property, Str, TraitDictEvent,
    TraitListEvent
)

# Logging.
logger = logging.getLogger(__name__)


class Bridge(HasTraits):
    """ Bridge that handles the client-server communication. """

    #### 'Bridge' protocol ####################################################

    def send_event(self, event):
        """ Send an event. """

        raise NotImplementedError


class Server(HasTraits):
    """ Server that serves a Jigna view. """

    #### 'Server' protocol ####################################################

    #: Base url for serving content.
    base_url = Property(Str, depends_on='_base_url')
    def _get_base_url(self):
        if not self._base_url.endswith('/'):
            base_url = self._base_url + '/'
        else:
            base_url = self._base_url

        return base_url

    def _set_base_url(self, base_url):
        self._base_url = base_url

    #: URL for the main home page
    home_url = Str('http://root.jigna/index.html')

    #: The html to serve.
    html = Str

    #: The trait change dispatch mechanism to use when traits change.
    trait_change_dispatch = Str('ui')

    #: Context mapping from object name to obj.
    context = Dict
    def _context_changed(self):
        self._register_objects(self.context)

    def _context_items_changed(self, dict_event):
        context = dict(dict_event.added)
        context.update(dict_event.changed)

        self._register_objects(context)

        self._send_context_updated_event(context)

        return

    def send_event(self, event):
        """ Send an event to the client(s). """

        self._bridge.send_event(event)

        return

    def handle_request(self, jsonized_request):
        """ Handle a jsonized request from a client. """

        request = json.loads(jsonized_request)

        # To dispatch the request we have a method named after each one!
        method    = getattr(self, request['kind'])
        exception = None
        try:
            result = method(request)

        except:
            exception = traceback.format_exc()
            logger.exception(exception)
            result = None

        response = dict(exception=exception, result=result)

        return json.dumps(response, default=lambda obj: repr(type(obj)));

    def shutdown(self):
        """ Shutdown the server.

        This unhooks the listeners to trait change events on any objects
        that we know about.

        """

        for obj in self._id_to_object_map.values():
            if isinstance(obj, HasTraits):
                obj.on_trait_change(
                    self._send_object_changed_event, remove=True
                )

    #### Handlers for each kind of request ####################################

    def update_context(self, request):
        """ Update the context on the JS side """
        # This method is called on a page reload or if a new client is used.
        # In these cases, the visited_type_names is no longer relevant as the
        # new client does not have the required data, so clear it.
        self._visited_type_names = set()
        return self._send_context_updated_event(self.context)

    def print_JS_message(self, request):
        """ Prints a message coming from the JS client for testing purposes """

        print('JS: ' + str(request['value']))

        return

    #### Instances ####

    def call_instance_method(self, request):
        """ Call a method on an instance. """

        obj         = self._id_to_object_map[request['id']]
        method_name = request['method_name']
        args        = self._unmarshal_all(request['args'])
        method      = getattr(obj, method_name)

        return self._marshal(method(*args))

    def call_instance_method_thread(self, request):
        """ Call a method on an instance *in a new thread*.

        Return the Id of a Future object which finishes when the method in
        thread finishes.

        """

        obj         = self._id_to_object_map[request['id']]
        method_name = request['method_name']
        args        = self._unmarshal_all(request['args'])
        method      = getattr(obj, method_name)

        from jigna.core.concurrent import Future
        future = Future(
            method, args=tuple(args), dispatch=self.trait_change_dispatch
        )

        def _on_done(result):
            event = dict(
                obj  = str(id(future)),
                name = 'done',
                data = result
            )
            self.send_event(event)

        def _on_error(error):
            error_msg = ''.join(traceback.format_exception(*error))

            logger.error(error_msg)

            event = dict(
                obj  = str(id(future)),
                name = 'error',
                data = error_msg
            )
            self.send_event(event)

        future.on_done(_on_done)
        future.on_error(_on_error)

        return self._marshal(id(future))

    def get_instance_attribute(self, request):
        """ Get the value of an instance attribute. """

        obj            = self._id_to_object_map[request['id']]
        attribute_name = request['attribute_name']

        return self._marshal(getattr(obj, attribute_name))

    def set_instance_attribute(self, request):
        """ Set an attribute on an instance. """

        obj            = self._id_to_object_map[request['id']]
        attribute_name = request['attribute_name']
        value          = self._unmarshal(request['value']);

        setattr(obj, attribute_name, value)

        return

    #### Lists/Dicts ####

    def get_item(self, request):
        """ Get the value of an item in a list or dict. """

        obj   = self._id_to_object_map[request['id']]
        index = request['index']

        return self._marshal(obj[index])

    def set_item(self, request):
        """ Set the value of a an item in a list or dict. """

        obj   = self._id_to_object_map[request['id']]
        index = request['index']
        value = self._unmarshal(request['value'])

        obj[index] = value

        return

    #### Private protocol #####################################################

    #: Shadow trait for `base_url`
    _base_url = Str

    #: The bridge that provides the communication between Python and JS.
    _bridge = Instance(Bridge)

    #: All instances, lists and dicts that have been accessed via the bridge.
    #:
    #: { str id : instance_list_or_dict obj }
    _id_to_object_map = Any
    def __id_to_object_map_default(self):
        return {}

    #: The typenames of the Python types that we have already visited.
    #:
    #: And by 'visited' we mean, those types that we have already sent the
    #: the full info for over to the client side.
    _visited_type_names = Any
    def __visited_type_names_default(self):
        return set()

    def _context_ids(self, context):
        """ Return a dictionary keyed with object ids of the objects in
        self._context and whose values are the object ids.
        """
        context_ids = {}
        for obj_name, obj in context.items():
            context_ids[obj_name] = self._marshal(obj)

        return context_ids

    def _get_attribute_names(self, obj):
        """ Get the names of all 'public' attributes on an object.

        Return a list of strings.

        """

        if isinstance(obj, HasTraits):
            attribute_names = [
                name for name in obj.editable_traits()

                if not name.startswith('_')
            ]
        else:
            attribute_names = [
                name for name, value in inspect.getmembers(obj)

                if not inspect.ismethod(value) and not name.startswith('_')
            ]

        return attribute_names

    def _get_dict_info(self, obj):
        """ Get a description of a dict. """

        return dict(keys=list(obj.keys()))

    def _get_event_names(self, obj):
        """ Get the names of all the attributes on an object.

        Return a list of strings.

        """

        # We ignore these triats that are added by the trait machinery :)
        ignore = ['trait_added', 'trait_modified']

        event_names = []

        if isinstance(obj, HasTraits):
            for trait_name in obj.class_trait_names():
                if obj.trait(trait_name).is_trait_type(Event):
                    if trait_name not in ignore:
                        event_names.append(trait_name)

        return event_names

    def _get_instance_info(self, obj):
        """ Get a description of an instance. """

        type_name = self._get_type_name(obj)

        # The first time we send the details of a type over to the client we
        # need to include the full info for it (its attributes, events and
        # methods etc)...
        if type_name not in self._visited_type_names:
            info = dict(
                type_name       = type_name,
                attribute_names = self._get_attribute_names(obj),
                event_names     = self._get_event_names(obj),
                method_names    = self._get_public_method_names(obj)
            )
            self._visited_type_names.add(type_name)

        # ... for subsequent calls, we only need to send the type name as the
        # the client will have already built a prototype based on the previous
        # info.
        else:
            info = dict(type_name = type_name)

        return info

    def _get_list_info(self, obj):
        """ Get a description of a list. """

        return dict(length=len(obj))

    def _get_public_method_names(self, obj):
        """ Get the names of all public methods on a class.

        Return a list of strings.

        """
        cls = type(obj)
        public_method_names = []
        for c in inspect.getmro(cls):
            if c is HasTraits:
                break

            for name in c.__dict__:
                if not name.startswith( '_' ):
                    value = getattr(obj, name)
                    if inspect.ismethod(value):
                        public_method_names.append(name)

        return public_method_names

    def _get_type_name(self, obj):
        t = type(obj)
        return t.__module__ + '.' + t.__name__

    def _marshal(self, obj):
        """ Marshal a value. """

        if isinstance(obj, list):
            obj_id = str(id(obj))
            self._id_to_object_map[obj_id] = obj

            type  = 'list'
            value = obj_id
            info  = self._get_list_info(obj)

        elif isinstance(obj, dict):
            obj_id = str(id(obj))
            self._id_to_object_map[obj_id] = obj

            type  = 'dict'
            value = obj_id
            info  = self._get_dict_info(obj)

        # fixme: Not quite right as this will be True for classes too ;^)
        # The intent is to get objects that are non-scalar eg. int, float
        # complex, str etc.
        elif hasattr(obj, '__dict__'):
            obj_id = str(id(obj))
            self._id_to_object_map[obj_id] = obj

            type  = 'instance'
            value = obj_id
            info  = self._get_instance_info(obj)

            if isinstance(obj, HasTraits):
                obj.on_trait_change(
                    self._send_object_changed_event,
                    dispatch=self.trait_change_dispatch
                )
        else:
            type  = 'primitive'
            value = obj
            info  = None

        return dict(type=type, value=value, info=info)

    def _marshal_all(self, iter):
        """ Marshal all of the values in an iterable. """

        return [self._marshal(obj) for obj in iter]

    def _unmarshal(self, obj):
        """ Unmarshal a value. """

        if obj['type'] == 'primitive':
            value = obj['value']

        else:
            value = self._id_to_object_map[obj['value']]

        return value

    def _unmarshal_all(self, iter):
        """ Unmarshal all of the values in an iterable. """

        return [self._unmarshal(obj) for obj in iter]

    def _register_object(self, obj):
        """ Register the given object with the server. """

        self._id_to_object_map[str(id(obj))] = obj

        return

    def _register_objects(self, objs):
        """ Register more than one objects """

        for obj_name, obj in objs.items():
            self._register_object(obj)

        return

    def _send_object_changed_event(self, obj, trait_name, old, new):
        """ Send an object changed event. """

        if trait_name.startswith('_'):
            return

        if isinstance(new, (TraitListEvent, TraitDictEvent)):
            trait_name  = trait_name[:-len('_items')]
            new         = getattr(obj, trait_name)
            items_event = True

        else:
            # fixme: intent is non-scalar or maybe container?
            if hasattr(new, '__dict__') or isinstance(new, (dict, list)):
                self._register_object(new)

            items_event = False

        event = dict(
            obj  = str(id(obj)),
            name = trait_name,
            # fixme: This smells a bit, but marshalling the new value gives us
            # a type/value pair which we need on the client side to determine
            # what (if any) proxy we need to create.
            data = self._marshal(new),

            # fixme: This is how we currently detect an 'xxx_items' event on
            # the JS side.
            items_event = items_event
        )

        self.send_event(event)

        return

    def _send_context_updated_event(self, context):
        """ Send a context_updated event. """

        event = dict(
            obj  = 'jigna',
            name = 'context_updated',
            data = self._context_ids(context)
        )

        self.send_event(event)

        return

#### EOF ######################################################################



from traits.api import Dict, HasTraits, Instance, Int, Str, List, Event
from jigna.template import Template

import unittest
import time


def sleep_while(condition, timeout, dt=0.1):
    def _check_cond():
        try:
            return condition()
        except Exception:
            return True

    t = 0.0
    while _check_cond():
        time.sleep(dt)
        t += dt
        if t > timeout:
            return False
    return True



#### Test model ####

class Person(HasTraits):
    name = Str
    age  = Int
    spouse = Instance('Person')
    fruits = List(Str)
    friends = List(Instance('Person'))
    phonebook = Dict(Str, Int)

    name_updated = Event

    def method(self, value):
        self.called_with = value

    def printme(self, val):
        print("JS: %s"%val)

    def method_slow(self, value, sleep_for):
        time.sleep(sleep_for)
        self.method_slow_called_with = value


class AddressBook(HasTraits):
    contacts = List

    def create(self):
        n = 10
        for i in range(n):
            self.contacts.append(Contact(name=str(i), number=str(i)))
        for i in range(n, 2*n):
            self.contacts.append(Company(
                name=str(i), number=str(i), address=str(i)
            ))


class Company(HasTraits):
    name = Str
    number = Str
    address = Str

class Contact(HasTraits):
    name = Str
    number = Str


#### UI for model ####

body_html = """
    <div>
      Name: <input ng-model="model.name">
      Age: <input ng-model="model.age" type='number'>
      <br/>
      Fruits:
      <ul>
        <li ng-repeat="fruit in model.fruits track by $index">
          <input ng-model="model.fruits[$index]">
        </li>
      </ul>

      <br/>

      Friends:
      <ul>
        <li ng-repeat="friend in model.friends">
          Name: <input ng-model="friend.name">
          Age: <input ng-model="friend.age" type="number">
          Fruits:
          <ul>
            <li ng-repeat="fruit in friend.fruits track by $index">
              <input ng-model="friend.fruits[$index]">
            </li>
          </ul>
        </li>
      </ul>
    </div>

    Spouse: <br/>
    Name: {{model.spouse.name}} Age: {{model.spouse.age}}

   <h3>Addressbook</h3>
   <button ng-click="jigna.threaded(addressbook, 'create')" id="create">
    Create
   </button>
   <br/>
   <ul>
    <li ng-repeat="contact in addressbook.contacts track by $index">
     <label>Name:</label> <input ng-model="contact.name">
     <label>Number:</label> <input ng-model="contact.number">
    </li>
   </ul>

"""

class TestJignaQt(unittest.TestCase):

    @classmethod
    def setUpClass(cls):
        from jigna.qt import QtGui
        from jigna.html_widget import HTMLWidget
        from jigna.utils import gui
        qapp = QtGui.QApplication.instance() or QtGui.QApplication([])
        template = Template(body_html=body_html)
        fred = Person(name='Fred', age=42)
        addressbook = AddressBook()
        widget = HTMLWidget(
            template=template,
            context={'model':fred, 'addressbook': addressbook}
        )
        widget.show()
        gui.process_events()
        cls.widget = widget
        cls.fred = fred
        cls.addressbook = addressbook

    def setUp(self):
        cls = self.__class__
        self.widget = cls.widget
        self.fred = cls.fred
        self.fred.spouse = None
        self.fred.fruits = []
        self.fred.friends = []
        self.fred.called_with = None
        self.fred.method_slow_called_with = None
        self.addressbook = cls.addressbook

    def wait_and_assert(self, condition, timeout=1.0):
        self.assertTrue(sleep_while(condition, timeout=timeout))

    def execute_js(self, js):
        from jigna.utils import gui
        gui.process_events()
        result = self.widget.execute_js(js)
        gui.process_events()
        return result

    def process_events(self):
        from jigna.utils import gui
        gui.process_events()

    def assertJSEqual(self, js, value):
        result = self.execute_js(js)
        if isinstance(value, (list, tuple)):
            msg = "Lengths different: expected %d, got %d" % \
                (len(value), len(result))
            self.assertEqual(len(value), len(result), msg)
            for index in range(len(value)):
                expect = value[index]
                got = result[str(index)]
                msg = "%s[%s] != %s, got %s"%(js, index, expect, got)
                self.assertEqual(expect, got, msg)
        else:
            msg = "%s != %s, got %s"%(js, value, result)
            self.assertEqual(value, result, msg)

    def test_object_identity(self):
        fred = self.fred
        wilma = Person(name='wilma')
        fred.friends.append(wilma)
        fred.spouse = wilma

        self.assertJSEqual(
            "jigna.models.model.friends[0] === jigna.models.model.spouse", True
        )

    def test_simple_primitive_traits(self):
        fred = self.fred
        fred.name = "Freddie"
        self.assertJSEqual("jigna.models.model.name", fred.name)
        fred.age = 43
        self.assertJSEqual("jigna.models.model.age", fred.age)

    def test_list_of_primitives(self):
        self.assertJSEqual("jigna.models.model.fruits", [])
        fred = self.fred
        fred.fruits = ["peach", "pear"]
        self.assertJSEqual("jigna.models.model.fruits", fred.fruits)

        fred.fruits.append('mango')
        self.assertJSEqual("jigna.models.model.fruits[0]", "peach")
        self.assertJSEqual("jigna.models.model.fruits[1]", "pear")
        self.assertJSEqual("jigna.models.model.fruits[2]", "mango")

        fred.fruits.insert(0, 'banana')
        self.assertJSEqual("jigna.models.model.fruits[0]", "banana")
        self.assertJSEqual("jigna.models.model.fruits[1]", "peach")
        self.assertJSEqual("jigna.models.model.fruits[2]", "pear")
        self.assertJSEqual("jigna.models.model.fruits[3]", "mango")

        # Test setting from the JS side...
        self.execute_js("jigna.models.model.fruits[0] = 'orange'")
        sleep_while(lambda: fred.fruits[0] != 'orange', timeout=1.0)
        self.assertEqual(fred.fruits, ["orange", "peach", "pear", "mango"])

        self.execute_js("jigna.models.model.fruits = ['apple']")
        sleep_while(lambda: fred.fruits[0] != 'apple', timeout=1.0)
        self.assertEqual(fred.fruits, ["apple"])

    def test_list_slicing(self):
        fred = self.fred
        fred.fruits = ["peach", "pear", "banana", "fruit", "apple"]
        self.assertJSEqual("jigna.models.model.fruits", fred.fruits)

        # Test that deleting a single element works.
        del fred.fruits[3]
        self.assertJSEqual("jigna.models.model.fruits", fred.fruits)

        # Now try a complex slice.

        # Check if negative strides work.
        fred.fruits[::-2] = ["mango", "litchi"]
        self.assertJSEqual("jigna.models.model.fruits", fred.fruits)

        # Check if positive strides work.
        fred.fruits[::2] = ["mango", "litchi"]
        self.assertJSEqual("jigna.models.model.fruits", fred.fruits)

        # Now try deleting a slice
        del fred.fruits[::2]
        self.assertJSEqual("jigna.models.model.fruits", fred.fruits)

        fred.fruits = ["peach", "pear", "banana", "fruit", "apple"]
        del fred.fruits[::3]
        self.assertJSEqual("jigna.models.model.fruits", fred.fruits)

    def test_dict_of_primitives(self):
        self.assertJSEqual("jigna.models.model.phonebook", {})
        fred = self.fred
        fred.phonebook = {'joe' : 123, 'joan' : 345}
        self.assertJSEqual("jigna.models.model.phonebook.joe", 123)
        self.assertJSEqual("jigna.models.model.phonebook.joan", 345)
        self.assertJSEqual("jigna.models.model.phonebook", fred.phonebook)

        fred.phonebook['wilma'] = 678
        self.assertJSEqual("jigna.models.model.phonebook.wilma", 678)
        self.assertJSEqual(
            "Object.keys(jigna.models.model.phonebook).length", 3
        )

        fred.phonebook['wilma'] = 890
        self.assertJSEqual("jigna.models.model.phonebook.wilma", 890)
        self.assertJSEqual(
            "Object.keys(jigna.models.model.phonebook).length", 3
        )

        fred.phonebook.pop('wilma')
        self.assertJSEqual("jigna.models.model.phonebook.wilma", None)
        self.assertJSEqual(
            "Object.keys(jigna.models.model.phonebook).length", 2
        )

        # Test setting from the JS side...
        self.execute_js("jigna.models.model.phonebook['joe'] = 567")
        sleep_while(lambda: fred.phonebook['joe'] != 567, timeout=1.0)
        self.assertEqual(567, fred.phonebook['joe'])

        self.execute_js("jigna.models.model.phonebook = {'alan' : 987}")
        sleep_while(lambda: fred.phonebook != {'alan': 987}, timeout=1.0)
        self.assertEqual(fred.phonebook, {'alan': 987})

    def test_instance_trait(self):
        self.assertIn(self.execute_js("jigna.models.model.spouse"), ['', None])
        wilma = Person(name='Wilma', age=40)
        self.fred.spouse = wilma
        self.assertJSEqual("jigna.models.model.spouse.name", 'Wilma')
        self.assertJSEqual("jigna.models.model.spouse.age", 40)

        # Set in the JS side.
        self.execute_js("jigna.models.model.spouse.name = 'Wilmaji'")
        self.execute_js("jigna.models.model.spouse.age = 41")
        sleep_while(lambda: wilma.age != 41, timeout=1.0)
        self.assertEqual(wilma.name, "Wilmaji")
        self.assertEqual(wilma.age, 41)

    def test_list_of_instances(self):
        self.assertJSEqual("jigna.models.model.friends", [])
        dino = Person(name="Dino", age=10)
        fred = self.fred
        fred.friends = [dino]
        self.assertJSEqual("jigna.models.model.friends[0].name", "Dino")
        self.assertJSEqual("jigna.models.model.friends[0].age", 10)

        wilma = Person(name="Wilma", age=30)
        fred.friends.append(wilma)
        self.assertJSEqual("jigna.models.model.friends[0].name", "Dino")
        self.assertJSEqual("jigna.models.model.friends[0].age", 10)
        self.assertJSEqual("jigna.models.model.friends[1].name", "Wilma")
        self.assertJSEqual("jigna.models.model.friends[1].age", 30)

        barney = Person(name="Barney", age=40)
        fred.friends.insert(0, barney)
        self.assertJSEqual("jigna.models.model.friends[0].name", "Barney")
        self.assertJSEqual("jigna.models.model.friends[0].age", 40)
        self.assertJSEqual("jigna.models.model.friends[1].name", "Dino")
        self.assertJSEqual("jigna.models.model.friends[1].age", 10)
        self.assertJSEqual("jigna.models.model.friends[2].name", "Wilma")
        self.assertJSEqual("jigna.models.model.friends[2].age", 30)

        # Test setting from the JS side...
        self.execute_js("jigna.models.model.friends[0].name = 'Barneyji'")
        sleep_while(lambda: barney.name != "Barneyji", timeout=1.0)
        self.assertEqual(barney.name, "Barneyji")

    def test_list_sortable(self):
        # Given
        fred = self.fred
        self.execute_js(
            "jigna.models.model.fruits = ['peach', 'apple', 'banana']"
        )
        self.assertJSEqual(
            "jigna.models.model.fruits", ['peach', 'apple', 'banana']
        )

        # When
        self.execute_js("jigna.models.model.fruits.sort()")

        # Then
        self.assertJSEqual(
            "jigna.models.model.fruits", ['apple', 'banana', 'peach']
        )

        sleep_while(lambda: fred.fruits[0] != 'apple', timeout=1.0)
        self.assertEqual(fred.fruits, ['apple', 'banana', 'peach'])

    def test_callable(self):
        fred = self.fred
        wilma = Person(name='Wilma', age=40)
        self.fred.spouse = wilma
        self.execute_js("jigna.models.model.method('hello')")
        self.assertEqual(fred.called_with, "hello")
        self.execute_js("jigna.models.model.method(1)")
        self.assertEqual(fred.called_with, 1)
        self.execute_js("jigna.models.model.method(10.0)")
        self.assertEqual(fred.called_with, 10.0)
        self.execute_js("jigna.models.model.method([1,2])")
        self.assertEqual(fred.called_with, [1,2])
        self.execute_js("jigna.models.model.method(jigna.models.model.spouse)")
        self.assertEqual(fred.called_with, wilma)

    def test_events_js(self):
        # When
        self.execute_js("""
            jigna.add_listener(jigna.models.model, 'name_updated', function(event){
                jigna.models.model.new_name = event.data.value;
            })
        """)
        self.fred.name_updated = "Freddie"

        # Then
        self.assertJSEqual("jigna.models.model.new_name", "Freddie")

    def test_threaded_call(self):
        # When
        self.execute_js("""
            var deferred = jigna.threaded(jigna.models.model, 'method_slow', 'foo', 1);
            deferred.done(function(){
                jigna.models.model.method_slow_finished = true;
            })
        """)

        # The following is a way to test async methods.
        # Basically, we keep on checking for a condition which becomes true
        # when the function is finished, and raise an error if the timeout
        # occurs before the condition becomes true
        timeout = 1.5
        dt = 0.1
        for i in range(int(timeout/dt)):
            time.sleep(dt)
            try:
                self.assertJSEqual('jigna.models.model.method_slow_finished', True)
                self.assertEqual(self.fred.method_slow_called_with, 'foo')
                break
            except Exception:
                pass
        else:
            raise AssertionError("Async method not finished")

    def test_nested_object_with_threaded_creation(self):
        # When

        # We trigger the click on setTimeout so as to avoid problems with
        # the JS code blocking any event processing
        self.execute_js(
            "setTimeout(function(){$('#create').trigger('click');});"
        )

        # Wait till the call is made.
        t = 0
        while len(self.addressbook.contacts) != 20 and t < 2.0:
            self.process_events()
            time.sleep(0.05)
            t += 0.05
        self.process_events()

        # Then.
        self.assertJSEqual('jigna.models.addressbook.contacts.length', 20)
        self.assertJSEqual('jigna.models.addressbook.contacts[1].name', '1')
        self.assertJSEqual('jigna.models.addressbook.contacts[1].number', '1')
        self.assertJSEqual('jigna.models.addressbook.contacts[10].name', '10')
        self.assertJSEqual('jigna.models.addressbook.contacts[10].number', '10')
        self.assertJSEqual('jigna.models.addressbook.contacts[10].address', '10')


if __name__ == "__main__":
    unittest.main()

#### EOF ######################################################################



"""
This example is the simplest example which shows how to start a jigna view.
"""

#### Imports ####
from __future__ import print_function

from jigna.api import HTMLWidget, Template
from jigna.qt import QtGui
from traits.api import HasTraits, Int, Str

#### Domain model ####

class Person(HasTraits):
    name = Str
    age  = Int

#### UI layer ####

body_html = """
    <div>
      Name: <input ng-model="person.name"><br>
      Age: <input ng-model="person.age" type='number'>
    </div>
"""

template = Template(body_html=body_html)

#### Entry point ####

def main():
    # Start a QtGui application
    app = QtGui.QApplication([])

    # Instantiate the domain model
    fred = Person(name='Fred', age=42)

    # Create a jigna based HTML widget to render the HTML template with the
    # given context.
    widget = HTMLWidget(template=template, context={'person': fred})
    widget.show()

    # Start the event loop
    app.exec_()

    # Check the values after the UI is closed
    print(fred.name, fred.age)

if __name__ == "__main__":
    main()

#### EOF ######################################################################



#
# (C) Copyright 2013-2017 Enthought, Inc., Austin, TX
# All right reserved.
#

import os.path
from setuptools import setup, find_packages

data = {}
fname = os.path.join('jigna', '__init__.py')
exec(compile(open(fname).read(), fname, 'exec'), data)

requires = ['traits', 'tornado']
pyside = requires + ['pyside']
test = pyside + ['nose', 'mock', 'coverage', 'selenium']

with open('README.rst') as fp:
    long_desc = fp.read()


setup(
    name='jigna',
    version=data['__version__'],
    author='Enthought, Inc',
    author_email='info@enthought.com',
    url='https://github.com/enthought/jigna',
    license='BSD',
    description='HTML frontend for Traits models.',
    long_description=long_desc,
    packages=find_packages(),
    package_dir={'jigna': 'jigna'},
    include_package_data=True,
    package_data={'jigna': ['js/dist/*.js']},
    install_requires=requires,
    extras_require={
        'pyside': pyside,
        'test': test,
    },
    classifiers=[
        c.strip() for c in """\
        Development Status :: 4 - Beta
        Environment :: Web Environment
        Intended Audience :: Developers
        Intended Audience :: Science/Research
        License :: OSI Approved :: BSD License
        Operating System :: MacOS :: MacOS X
        Operating System :: Microsoft :: Windows
        Operating System :: OS Independent
        Operating System :: POSIX
        Operating System :: Unix
        Programming Language :: Python
        Programming Language :: JavaScript
        Topic :: Software Development
        Topic :: Software Development :: Libraries
        Topic :: Software Development :: User Interfaces
        """.splitlines() if len(c.split()) > 0
    ],
    zip_safe=False
)
